# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
This module holds the standard implementation of the :class:`PrinterInterface` and it helpers.
"""

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import copy
import logging
import os
import threading
import time

from past.builtins import basestring, long

from frozendict import frozendict

from octoprint import util as util
from octoprint.events import eventManager, Events
from octoprint.filemanager import FileDestinations, NoSuchStorage, valid_file_type
from octoprint.plugin import plugin_manager, ProgressPlugin
from octoprint.printer import PrinterInterface, PrinterCallback, UnknownScript, InvalidFileLocation, InvalidFileType
from octoprint.printer.estimation import PrintTimeEstimator
from octoprint.settings import settings
from octoprint.util import comm as comm
from octoprint.util import InvariantContainer
from octoprint.util import to_unicode
from octoprint.util import monotonic_time
from octoprint.util import get_fully_qualified_classname as fqcn


class Printer(PrinterInterface, comm.MachineComPrintCallback):
	"""
	Default implementation of the :class:`PrinterInterface`. Manages the communication layer object and registers
	itself with it as a callback to react to changes on the communication layer.
	"""

	def __init__(self, fileManager, analysisQueue, printerProfileManager):
		from collections import deque

		self._logger = logging.getLogger(__name__)
		self._logger_job = logging.getLogger("{}.job".format(__name__))

		self._dict = frozendict if settings().getBoolean(["devel", "useFrozenDictForPrinterState"]) else dict

		self._analysisQueue = analysisQueue
		self._fileManager = fileManager
		self._printerProfileManager = printerProfileManager

		# state
		# TODO do we really need to hold the temperature here?
		self._temp = None
		self._bedTemp = None
		self._chamberTemp = None
		self._targetTemp = None
		self._targetBedTemp = None
		self._targetChamberTemp = None
		self._temps = TemperatureHistory(cutoff=settings().getInt(["temperature", "cutoff"])*60)
		self._tempBacklog = []

		self._messages = deque([], 300)
		self._messageBacklog = []

		self._log = deque([], 300)
		self._logBacklog = []

		self._state = None

		self._currentZ = None

		self._printAfterSelect = False
		self._posAfterSelect = None

		# sd handling
		self._sdPrinting = False
		self._sdStreaming = False
		self._sdFilelistAvailable = threading.Event()
		self._streamingFinishedCallback = None
		self._streamingFailedCallback = None

		# job handling & estimation
		self._selectedFileMutex = threading.RLock()
		self._selectedFile = None

		self._estimator_factory = PrintTimeEstimator
		self._estimator = None
		analysis_queue_hooks = plugin_manager().get_hooks("octoprint.printer.estimation.factory")
		for name, hook in analysis_queue_hooks.items():
			try:
				estimator = hook()
				if estimator is not None:
					self._logger.info("Using print time estimator provided by {}".format(name))
					self._estimator_factory = estimator
			except Exception:
				self._logger.exception("Error while processing analysis queues from {}".format(name),
				                       extra=dict(plugin=name))

		#hook card upload
		self.sd_card_upload_hooks = plugin_manager().get_hooks("octoprint.printer.sdcardupload")

		# comm
		self._comm = None

		# callbacks
		self._callbacks = []

		# progress plugins
		self._lastProgressReport = None
		self._progressPlugins = plugin_manager().get_implementations(ProgressPlugin)

		self._stateMonitor = StateMonitor(
			interval=0.5,
			on_update=self._sendCurrentDataCallbacks,
			on_add_temperature=self._sendAddTemperatureCallbacks,
			on_add_log=self._sendAddLogCallbacks,
			on_add_message=self._sendAddMessageCallbacks,
			on_get_progress=self._updateProgressDataCallback
		)
		self._stateMonitor.reset(
			state=self._dict(text=self.get_state_string(), flags=self._getStateFlags()),
			job_data=self._dict(file=self._dict(name=None,
			                                    path=None,
			                                    size=None,
			                                    origin=None,
			                                    date=None),
			                    estimatedPrintTime=None,
			                    lastPrintTime=None,
			                    filament=self._dict(length=None,
			                                        volume=None),
			                    user=None),
			progress=self._dict(completion=None,
			                    filepos=None,
			                    printTime=None,
			                    printTimeLeft=None,
			                    printTimeOrigin=None),
			current_z=None,
			offsets=self._dict()
		)

		eventManager().subscribe(Events.METADATA_ANALYSIS_FINISHED, self._on_event_MetadataAnalysisFinished)
		eventManager().subscribe(Events.METADATA_STATISTICS_UPDATED, self._on_event_MetadataStatisticsUpdated)

	def _create_estimator(self, job_type=None):
		if job_type is None:
			with self._selectedFileMutex:
				if self._selectedFile is None:
					return

				if self._selectedFile["sd"]:
					job_type = "sdcard"
				else:
					job_type = "local"

		self._estimator = self._estimator_factory(job_type)

	#~~ handling of PrinterCallbacks

	def register_callback(self, callback, *args, **kwargs):
		if not isinstance(callback, PrinterCallback):
			self._logger.warning("Registering an object as printer callback which doesn't implement the PrinterCallback interface")
		self._callbacks.append(callback)

	def unregister_callback(self, callback, *args, **kwargs):
		try:
			self._callbacks.remove(callback)
		except ValueError:
			# not registered
			pass

	def send_initial_callback(self, callback):
		if callback in self._callbacks:
			self._sendInitialStateUpdate(callback)

	def _sendAddTemperatureCallbacks(self, data):
		for callback in self._callbacks:
			try:
				callback.on_printer_add_temperature(data)
			except Exception:
				self._logger.exception("Exception while adding temperature data point to callback {}".format(callback),
				                       extra=dict(callback=fqcn(callback)))

	def _sendAddLogCallbacks(self, data):
		for callback in self._callbacks:
			try:
				callback.on_printer_add_log(data)
			except Exception:
				self._logger.exception("Exception while adding communication log entry to callback {}".format(callback),
				                       extra=dict(callback=fqcn(callback)))

	def _sendAddMessageCallbacks(self, data):
		for callback in self._callbacks:
			try:
				callback.on_printer_add_message(data)
			except Exception:
				self._logger.exception("Exception while adding printer message to callback {}".format(callback),
				                       extra=dict(callback=fqcn(callback)))

	def _sendCurrentDataCallbacks(self, data):
		for callback in self._callbacks:
			try:
				callback.on_printer_send_current_data(copy.deepcopy(data))
			except Exception:
				self._logger.exception("Exception while pushing current data to callback {}".format(callback),
				                       extra=dict(callback=fqcn(callback)))

	#~~ callback from metadata analysis event

	def _on_event_MetadataAnalysisFinished(self, event, data):
		with self._selectedFileMutex:
			if self._selectedFile:
				self._setJobData(self._selectedFile["filename"],
								 self._selectedFile["filesize"],
								 self._selectedFile["sd"],
								 self._selectedFile["user"])

	def _on_event_MetadataStatisticsUpdated(self, event, data):
		with self._selectedFileMutex:
			if self._selectedFile:
				self._setJobData(self._selectedFile["filename"],
				                 self._selectedFile["filesize"],
				                 self._selectedFile["sd"],
				                 self._selectedFile["user"])

	#~~ progress plugin reporting

	def _reportPrintProgressToPlugins(self, progress):
		with self._selectedFileMutex:
			if progress is None or not self._selectedFile or not "sd" in self._selectedFile or not "filename" in self._selectedFile:
				return

			storage = "sdcard" if self._selectedFile["sd"] else "local"
			filename = self._selectedFile["filename"]

		def call_plugins(storage, filename, progress):
			for plugin in self._progressPlugins:
				try:
					plugin.on_print_progress(storage, filename, progress)
				except Exception:
					self._logger.exception("Exception while sending print progress to plugin %s" % plugin._identifier,
					                       extra=dict(plugin=plugin._identifier))

		thread = threading.Thread(target=call_plugins, args=(storage, filename, progress))
		thread.daemon = False
		thread.start()

	#~~ PrinterInterface implementation

	def connect(self, port=None, baudrate=None, profile=None, *args, **kwargs):
		"""
		 Connects to the printer. If port and/or baudrate is provided, uses these settings, otherwise autodetection
		 will be attempted.
		"""
		if self._comm is not None:
			self.disconnect()

		eventManager().fire(Events.CONNECTING)
		self._printerProfileManager.select(profile)

		from octoprint.logging.handlers import SerialLogHandler
		SerialLogHandler.on_open_connection()
		if not logging.getLogger("SERIAL").isEnabledFor(logging.DEBUG):
			# if serial.log is not enabled, log a line to explain that to reduce "serial.log is empty" in tickets...
			logging.getLogger("SERIAL").info("serial.log is currently not enabled, you can enable it via Settings > Serial Connection > Log communication to serial.log")

		self._comm = comm.MachineCom(port, baudrate, callbackObject=self, printerProfileManager=self._printerProfileManager)

	def disconnect(self, *args, **kwargs):
		"""
		 Closes the connection to the printer.
		"""
		eventManager().fire(Events.DISCONNECTING)
		if self._comm is not None:
			self._comm.close()
		else:
			eventManager().fire(Events.DISCONNECTED)

	def get_transport(self, *args, **kwargs):

		if self._comm is None:
			return None

		return self._comm.getTransport()
	getTransport = util.deprecated("getTransport has been renamed to get_transport", since="1.2.0-dev-590", includedoc="Replaced by :func:`get_transport`")

	def job_on_hold(self, blocking=True, *args, **kwargs):
		if self._comm is None:
			raise RuntimeError("No connection to the printer")
		return self._comm.job_put_on_hold(blocking=blocking)

	def set_job_on_hold(self, value, blocking=True, *args, **kwargs):
		if self._comm is None:
			raise RuntimeError("No connection to the printer")
		return self._comm.set_job_on_hold(value, blocking=blocking)

	def fake_ack(self, *args, **kwargs):
		if self._comm is None:
			return

		self._comm.fakeOk()

	def commands(self, commands, *args, **kwargs):
		"""
		Sends one or more gcode commands to the printer.
		"""
		if self._comm is None:
			return

		if not isinstance(commands, (list, tuple)):
			commands = [commands]

		for command in commands:
			self._comm.sendCommand(command, tags=kwargs.get("tags", set()) | {"trigger:printer.commands"})

	def script(self, name, context=None, must_be_set=True, part_of_job=False, *args, **kwargs):
		if self._comm is None:
			return

		if name is None or not name:
			raise ValueError("name must be set")

		result = self._comm.sendGcodeScript(name,
		                                    part_of_job=part_of_job,
		                                    replacements=context,
		                                    tags=kwargs.get("tags", set()) | {"trigger:printer.script"})
		if not result and must_be_set:
			raise UnknownScript(name)

	def jog(self, axes, relative=True, speed=None, *args, **kwargs):
		if isinstance(axes, basestring):
			# legacy parameter format, there should be an amount as first anonymous positional arguments too
			axis = axes

			if not len(args) >= 1:
				raise ValueError("amount not set")
			amount = args[0]
			if not isinstance(amount, (int, long, float)):
				raise ValueError("amount must be a valid number: {amount}".format(amount=amount))

			axes = dict()
			axes[axis] = amount

		if not axes:
			raise ValueError("At least one axis to jog must be provided")

		for axis in axes:
			if not axis in PrinterInterface.valid_axes:
				raise ValueError("Invalid axis {}, valid axes are {}".format(axis, ", ".join(PrinterInterface.valid_axes)))

		command = "G1 {}".format(" ".join(["{}{}".format(axis.upper(), amt) for axis, amt in axes.items()]))

		if speed is None:
			printer_profile = self._printerProfileManager.get_current_or_default()
			speed = min([printer_profile["axes"][axis]["speed"] for axis in axes])

		if speed and not isinstance(speed, bool):
			command += " F{}".format(speed)

		if relative:
			commands = ["G91", command, "G90"]
		else:
			commands = ["G90", command]

		self.commands(commands, tags=kwargs.get("tags", set()) | {"trigger:printer.jog"})

	def home(self, axes, *args, **kwargs):
		if not isinstance(axes, (list, tuple)):
			if isinstance(axes, basestring):
				axes = [axes]
			else:
				raise ValueError("axes is neither a list nor a string: {axes}".format(axes=axes))

		validated_axes = list(filter(lambda x: x in PrinterInterface.valid_axes, map(lambda x: x.lower(), axes)))
		if len(axes) != len(validated_axes):
			raise ValueError("axes contains invalid axes: {axes}".format(axes=axes))

		self.commands(["G91", "G28 %s" % " ".join(map(lambda x: "%s0" % x.upper(), validated_axes)), "G90"],
		              tags=kwargs.get("tags", set) | {"trigger:printer.home"})

	def extrude(self, amount, speed=None, *args, **kwargs):
		if not isinstance(amount, (int, long, float)):
			raise ValueError("amount must be a valid number: {amount}".format(amount=amount))

		printer_profile = self._printerProfileManager.get_current_or_default()

		# Use specified speed (if any)
		max_e_speed = printer_profile["axes"]["e"]["speed"]

		if speed is None:
			# No speed was specified so default to value configured in printer profile
			extrusion_speed = max_e_speed
		else:
			# Make sure that specified value is not greater than maximum as defined in printer profile
			extrusion_speed = min([speed, max_e_speed])

		self.commands(["G91", "G1 E%s F%d" % (amount, extrusion_speed), "G90"],
		              tags=kwargs.get("tags", set()) | {"trigger:printer.extrude"})

	def change_tool(self, tool, *args, **kwargs):
		if not PrinterInterface.valid_tool_regex.match(tool):
			raise ValueError("tool must match \"tool[0-9]+\": {tool}".format(tool=tool))

		tool_num = int(tool[len("tool"):])
		self.commands("T%d" % tool_num, tags=kwargs.get("tags", set()) | {"trigger:printer.change_tool"})

	def set_temperature(self, heater, value, *args, **kwargs):
		if not PrinterInterface.valid_heater_regex.match(heater):
			raise ValueError("heater must match \"tool[0-9]+\", \"bed\" or \"chamber\": {heater}".format(heater=heater))

		if not isinstance(value, (int, long, float)) or value < 0:
			raise ValueError("value must be a valid number >= 0: {value}".format(value=value))

		tags = kwargs.get("tags", set()) | {"trigger:printer.set_temperature"}

		if heater.startswith("tool"):
			printer_profile = self._printerProfileManager.get_current_or_default()
			extruder_count = printer_profile["extruder"]["count"]
			shared_nozzle = printer_profile["extruder"]["sharedNozzle"]
			if extruder_count > 1 and not shared_nozzle:
				toolNum = int(heater[len("tool"):])
				self.commands("M104 T{} S{}".format(toolNum, value),
				              tags=tags)
			else:
				self.commands("M104 S{}".format(value),
				              tags=tags)

		elif heater == "bed":
			self.commands("M140 S{}".format(value),
			              tags=tags)

		elif heater == "chamber":
			self.commands("M141 S{}".format(value),
			              tags=tags)

	def set_temperature_offset(self, offsets=None, *args, **kwargs):
		if offsets is None:
			offsets = dict()

		if not isinstance(offsets, dict):
			raise ValueError("offsets must be a dict")

		validated_keys = list(filter(lambda x: PrinterInterface.valid_heater_regex.match(x), offsets.keys()))
		validated_values = list(filter(lambda x: isinstance(x, (int, long, float)), offsets.values()))

		if len(validated_keys) != len(offsets):
			raise ValueError("offsets contains invalid keys: {offsets}".format(offsets=offsets))
		if len(validated_values) != len(offsets):
			raise ValueError("offsets contains invalid values: {offsets}".format(offsets=offsets))

		if self._comm is None:
			return

		self._comm.setTemperatureOffset(offsets)
		self._setOffsets(self._comm.getOffsets())

	def _convert_rate_value(self, factor, min=0, max=200):
		if not isinstance(factor, (int, float, long)):
			raise ValueError("factor is not a number")

		if isinstance(factor, float):
			factor = int(factor * 100.0)

		if factor < min or factor > max:
			raise ValueError("factor must be a value between {} and {}".format(min, max))

		return factor

	def feed_rate(self, factor, *args, **kwargs):
		factor = self._convert_rate_value(factor, min=50, max=200)
		self.commands("M220 S%d" % factor,
		              tags=kwargs.get("tags", set()) | {"trigger:printer.feed_rate"})

	def flow_rate(self, factor, *args, **kwargs):
		factor = self._convert_rate_value(factor, min=75, max=125)
		self.commands("M221 S%d" % factor,
		              tags=kwargs.get("tags", set()) | {"trigger:printer.flow_rate"})

	def select_file(self, path, sd, printAfterSelect=False, user=None, pos=None, *args, **kwargs):
		if self._comm is None or (self._comm.isBusy() or self._comm.isStreaming()):
			self._logger.info("Cannot load file: printer not connected or currently busy")
			return

		self._validateJob(path, sd)

		origin = FileDestinations.SDCARD if sd else FileDestinations.LOCAL
		if sd:
			path_on_disk = "/" + path
			path_in_storage = path
		else:
			path_on_disk = self._fileManager.path_on_disk(origin, path)
			path_in_storage = self._fileManager.path_in_storage(origin, path_on_disk)

		try:
			recovery_data = self._fileManager.get_recovery_data()
			if recovery_data:
				# clean up recovery data if we just selected a different file
				actual_origin = recovery_data.get("origin", None)
				actual_path = recovery_data.get("path", None)

				if actual_origin is None or actual_path is None or actual_origin != origin or actual_path != path_in_storage:
					self._fileManager.delete_recovery_data()
		except Exception:
			# anything goes wrong with the recovery data, we ignore it
			self._logger.exception("Something was wrong with processing the recovery data")

		self._printAfterSelect = printAfterSelect
		self._posAfterSelect = pos
		self._comm.selectFile("/" + path if sd else path_on_disk, sd,
		                      user=user,
		                      tags=kwargs.get("tags", set()) | {"trigger:printer.select_file"})
		self._updateProgressData()
		self._setCurrentZ(None)

	def unselect_file(self, *args, **kwargs):
		if self._comm is not None and (self._comm.isBusy() or self._comm.isStreaming()):
			return

		self._comm.unselectFile()
		self._updateProgressData()
		self._setCurrentZ(None)

	def get_file_position(self):
		if self._comm is None:
			return None

		with self._selectedFileMutex:
			if self._selectedFile is None:
				return None

		return self._comm.getFilePosition()

	def start_print(self, pos=None, user=None, *args, **kwargs):
		"""
		 Starts the currently loaded print job.
		 Only starts if the printer is connected and operational, not currently printing and a printjob is loaded
		"""
		if self._comm is None or not self._comm.isOperational() or self._comm.isPrinting():
			return

		with self._selectedFileMutex:
			if self._selectedFile is None:
				return

		self._fileManager.delete_recovery_data()

		self._lastProgressReport = None
		self._updateProgressData()
		self._setCurrentZ(None)
		self._comm.startPrint(pos=pos,
		                      user=user,
		                      tags=kwargs.get("tags", set()) | {"trigger:printer.start_print"})

	def pause_print(self, user=None, *args, **kwargs):
		"""
		Pause the current printjob.
		"""
		if self._comm is None:
			return

		if self._comm.isPaused():
			return

		self._comm.setPause(True,
		                    user=user,
		                    tags=kwargs.get("tags", set()) | {"trigger:printer.pause_print"})

	def resume_print(self, user=None, *args, **kwargs):
		"""
		Resume the current printjob.
		"""
		if self._comm is None:
			return

		if not self._comm.isPaused():
			return

		self._comm.setPause(False,
		                    user=user,
		                    tags=kwargs.get("tags", set()) | {"trigger:printer.resume_print"})

	def cancel_print(self, user=None, *args, **kwargs):
		"""
		 Cancel the current printjob.
		"""
		if self._comm is None:
			return

		# tell comm layer to cancel - will also trigger our cancelled handler
		# for further processing
		self._comm.cancelPrint(user=user,
		                       tags=kwargs.get("tags", set()) | {"trigger:printer.cancel_print"})

	def log_lines(self, *lines):
		serial_logger = logging.getLogger("SERIAL")
		self.on_comm_log("\n".join(lines))
		for line in lines:
			serial_logger.debug(line)

	def get_state_string(self, state=None, *args, **kwargs):
		if self._comm is None:
			return "Offline"
		else:
			return self._comm.getStateString(state=state)

	def get_state_id(self, state=None, *args, **kwargs):
		if self._comm is None:
			return "OFFLINE"
		else:
			return self._comm.getStateId(state=state)

	def get_current_data(self, *args, **kwargs):
		return util.thaw_frozendict(self._stateMonitor.get_current_data())

	def get_current_job(self, *args, **kwargs):
		currentData = self._stateMonitor.get_current_data()
		return util.thaw_frozendict(currentData["job"])

	def get_current_temperatures(self, *args, **kwargs):
		if self._comm is not None:
			offsets = self._comm.getOffsets()
		else:
			offsets = dict()

		result = {}
		if self._temp is not None:
			for tool in self._temp.keys():
				result["tool%d" % tool] = {
					"actual": self._temp[tool][0],
					"target": self._temp[tool][1],
					"offset": offsets[tool] if tool in offsets and offsets[tool] is not None else 0
				}
		if self._bedTemp is not None:
			result["bed"] = {
				"actual": self._bedTemp[0],
				"target": self._bedTemp[1],
				"offset": offsets["bed"] if "bed" in offsets and offsets["bed"] is not None else 0
			}
		if self._chamberTemp is not None:
			result["chamber"] = {
				"actual": self._chamberTemp[0],
				"target": self._chamberTemp[1],
				"offset": offsets["chamber"] if "chamber" in offsets and offsets["chamber"] is not None else 0
			}

		return result

	def get_temperature_history(self, *args, **kwargs):
		return self._temps

	def get_current_connection(self, *args, **kwargs):
		if self._comm is None:
			return "Closed", None, None, None

		port, baudrate = self._comm.getConnection()
		printer_profile = self._printerProfileManager.get_current_or_default()
		return self._comm.getStateString(), port, baudrate, printer_profile

	def is_closed_or_error(self, *args, **kwargs):
		return self._comm is None or self._comm.isClosedOrError()

	def is_operational(self, *args, **kwargs):
		return self._comm is not None and self._comm.isOperational()

	def is_printing(self, *args, **kwargs):
		return self._comm is not None and self._comm.isPrinting()

	def is_cancelling(self, *args, **kwargs):
		return self._comm is not None and self._comm.isCancelling()

	def is_pausing(self, *args, **kwargs):
		return self._comm is not None and self._comm.isPausing()

	def is_paused(self, *args, **kwargs):
		return self._comm is not None and self._comm.isPaused()

	def is_resuming(self, *args, **kwargs):
		return self._comm is not None and self._comm.isResuming()

	def is_finishing(self, *args, **kwargs):
		return self._comm is not None and self._comm.isFinishing()

	def is_error(self, *args, **kwargs):
		return self._comm is not None and self._comm.isError()

	def is_ready(self, *args, **kwargs):
		return self.is_operational() and not self.is_printing() and not self._comm.isStreaming()

	def is_sd_ready(self, *args, **kwargs):
		if not settings().getBoolean(["feature", "sdSupport"]) or self._comm is None:
			return False
		else:
			return self._comm.isSdReady()

	#~~ sd file handling

	def get_sd_files(self, *args, **kwargs):
		if self._comm is None or not self._comm.isSdReady():
			return []
		return list(map(lambda x: (x[0][1:], x[1]), self._comm.getSdFiles()))

	def add_sd_file(self, filename, path, on_success=None, on_failure=None, *args, **kwargs):
		if not self._comm or self._comm.isBusy() or not self._comm.isSdReady():
			self._logger.error("No connection to printer or printer is busy")
			return

		self._streamingFinishedCallback = on_success
		self._streamingFailedCallback = on_failure

		def sd_upload_started(local_filename, remote_filename):
			eventManager().fire(Events.TRANSFER_STARTED, dict(local=local_filename,
			                                                  remote=remote_filename))

		def sd_upload_succeeded(local_filename, remote_filename, elapsed):
			payload = dict(local=local_filename,
			               remote=remote_filename,
			               time=elapsed)
			eventManager().fire(Events.TRANSFER_DONE, payload)
			if callable(self._streamingFinishedCallback):
				self._streamingFinishedCallback(remote_filename, remote_filename, FileDestinations.SDCARD)

		def sd_upload_failed(local_filename, remote_filename, elapsed):
			payload = dict(local=local_filename,
			               remote=remote_filename,
			               time=elapsed)
			eventManager().fire(Events.TRANSFER_FAILED, payload)
			if callable(self._streamingFailedCallback):
				self._streamingFailedCallback(remote_filename, remote_filename, FileDestinations.SDCARD)

		for name, hook in self.sd_card_upload_hooks.items():
			# first sd card upload plugin that feels responsible gets the job
			try:
				result = hook(self, filename, path, sd_upload_started, sd_upload_succeeded, sd_upload_failed,
				              *args, **kwargs)
				if result is not None:
					return result
			except Exception:
				self._logger.exception("There was an error running the sd upload "
				                       "hook provided by plugin {}".format(name),
				                       extra=dict(plugin=name))

		else:
			# no plugin feels responsible, use the default implementation
			return self._add_sd_file(filename, path, tags=kwargs.get("tags"))

	def _get_free_remote_name(self, filename):
		self.refresh_sd_files(blocking=True)
		existingSdFiles = list(map(lambda x: x[0], self._comm.getSdFiles()))

		if valid_file_type(filename, "gcode"):
			# figure out remote filename
			remote_name = util.get_dos_filename(filename,
			                                    existing_filenames=existingSdFiles,
			                                    extension="gco",
			                                    whitelisted_extensions=["gco", "g"])
		else:
			# probably something else added through a plugin, use it's basename as-is
			remote_name = os.path.basename(filename)

		return remote_name

	def _add_sd_file(self, filename, path, tags=None):
		if tags is None:
			tags = set()

		remote_name = self._get_free_remote_name(filename)
		self._create_estimator("stream")
		self._comm.startFileTransfer(path, filename, "/" + remote_name,
		                             special=not valid_file_type(filename, "gcode"),
		                             tags=tags | {"trigger:printer.add_sd_file"})

		return remote_name

	def delete_sd_file(self, filename, *args, **kwargs):
		if not self._comm or not self._comm.isSdReady():
			return
		self._comm.deleteSdFile("/" + filename, tags=kwargs.get("tags", set()) | {"trigger:printer.delete_sd_file"})

	def init_sd_card(self, *args, **kwargs):
		if not self._comm or self._comm.isSdReady():
			return
		self._comm.initSdCard(tags=kwargs.get("tags", set()) | {"trigger:printer.init_sd_card"})

	def release_sd_card(self, *args, **kwargs):
		if not self._comm or not self._comm.isSdReady():
			return
		self._comm.releaseSdCard(tags=kwargs.get("tags", set()) | {"trigger:printer.release_sd_card"})

	def refresh_sd_files(self, blocking=False, *args, **kwargs):
		"""
		Refreshes the list of file stored on the SD card attached to printer (if available and printer communication
		available). Optional blocking parameter allows making the method block (max 10s) until the file list has been
		received (and can be accessed via self._comm.getSdFiles()). Defaults to an asynchronous operation.
		"""
		if not self._comm or not self._comm.isSdReady():
			return
		self._sdFilelistAvailable.clear()
		self._comm.refreshSdFiles(tags=kwargs.get("tags", set()) | {"trigger:printer.refresh_sd_files"})
		if blocking:
			self._sdFilelistAvailable.wait(kwargs.get("timeout", 10000))

	#~~ state monitoring

	def _setOffsets(self, offsets):
		self._stateMonitor.set_temp_offsets(offsets)

	def _setCurrentZ(self, currentZ):
		self._currentZ = currentZ
		self._stateMonitor.set_current_z(self._currentZ)

	def _setState(self, state, state_string=None):
		if state_string is None:
			state_string = self.get_state_string()

		self._state = state
		self._stateMonitor.set_state(self._dict(text=state_string, flags=self._getStateFlags()))

		payload = dict(
			state_id=self.get_state_id(self._state),
			state_string=self.get_state_string(self._state)
		)
		eventManager().fire(Events.PRINTER_STATE_CHANGED, payload)

	def _addLog(self, log):
		self._log.append(log)
		self._stateMonitor.add_log(log)

	def _addMessage(self, message):
		self._messages.append(message)
		self._stateMonitor.add_message(message)

	def _updateProgressData(self, completion=None, filepos=None, printTime=None, printTimeLeft=None, printTimeLeftOrigin=None):
		self._stateMonitor.set_progress(self._dict(completion=int(completion * 100) if completion is not None else None,
		                                           filepos=filepos,
		                                           printTime=int(printTime) if printTime is not None else None,
		                                           printTimeLeft=int(printTimeLeft) if printTimeLeft is not None else None,
		                                           printTimeLeftOrigin=printTimeLeftOrigin))

	def _updateProgressDataCallback(self):
		if self._comm is None:
			progress = None
			filepos = None
			printTime = None
			cleanedPrintTime = None
		else:
			progress = self._comm.getPrintProgress()
			filepos = self._comm.getPrintFilepos()
			printTime = self._comm.getPrintTime()
			cleanedPrintTime = self._comm.getCleanedPrintTime()

		printTimeLeft = printTimeLeftOrigin = None
		estimator = self._estimator
		if progress is not None:
			progress_int = int(progress * 100)
			if self._lastProgressReport != progress_int:
				self._lastProgressReport = progress_int
				self._reportPrintProgressToPlugins(progress_int)

			if progress == 0:
				printTimeLeft = None
				printTimeLeftOrigin = None
			elif progress == 1.0:
				printTimeLeft = 0
				printTimeLeftOrigin = None
			elif estimator is not None:
				statisticalTotalPrintTime = None
				statisticalTotalPrintTimeType = None
				with self._selectedFileMutex:
					if self._selectedFile and "estimatedPrintTime" in self._selectedFile \
							and self._selectedFile["estimatedPrintTime"]:
						statisticalTotalPrintTime = self._selectedFile["estimatedPrintTime"]
						statisticalTotalPrintTimeType = self._selectedFile.get("estimatedPrintTimeType", None)

				try:
					printTimeLeft, printTimeLeftOrigin = estimator.estimate(progress,
					                                                        printTime,
					                                                        cleanedPrintTime,
					                                                        statisticalTotalPrintTime,
					                                                        statisticalTotalPrintTimeType)
				except Exception:
					self._logger.exception("Error while estimating print time via {}".format(estimator))

		return self._dict(completion=progress * 100 if progress is not None else None,
		                  filepos=filepos,
		                  printTime=int(printTime) if printTime is not None else None,
		                  printTimeLeft=int(printTimeLeft) if printTimeLeft is not None else None,
		                  printTimeLeftOrigin=printTimeLeftOrigin)

	def _addTemperatureData(self, tools=None, bed=None, chamber=None):
		if tools is None:
			tools = dict()

		data = dict(time=int(time.time()))
		for tool in tools.keys():
			data["tool%d" % tool] = self._dict(actual=tools[tool][0], target=tools[tool][1])
		if bed is not None and isinstance(bed, tuple):
			data["bed"] = self._dict(actual=bed[0], target=bed[1])
		if chamber is not None and isinstance(chamber, tuple):
			data["chamber"] = self._dict(actual=chamber[0], target=chamber[1])

		self._temps.append(data)

		self._temp = tools
		self._bedTemp = bed
		self._chamberTemp = chamber

		self._stateMonitor.add_temperature(self._dict(**data))

	def _validateJob(self, filename, sd):
		if not valid_file_type(filename, type="machinecode"):
			raise InvalidFileType("{} is not a machinecode file, cannot print".format(filename))

		if sd:
			return

		path_on_disk = self._fileManager.path_on_disk(FileDestinations.LOCAL, filename)
		if os.path.isabs(filename) and not filename == path_on_disk:
			raise InvalidFileLocation("{} is not located within local storage, cannot select for printing".format(filename))
		if not os.path.isfile(path_on_disk):
			raise InvalidFileLocation("{} does not exist in local storage, cannot select for printing".format(filename))

	def _setJobData(self, filename, filesize, sd, user=None):
		with self._selectedFileMutex:
			if filename is not None:
				if sd:
					name_in_storage = filename
					if name_in_storage.startswith("/"):
						name_in_storage = name_in_storage[1:]
					path_in_storage = name_in_storage
					path_on_disk = None
				else:
					path_in_storage = self._fileManager.path_in_storage(FileDestinations.LOCAL, filename)
					path_on_disk = self._fileManager.path_on_disk(FileDestinations.LOCAL, filename)
					_, name_in_storage = self._fileManager.split_path(FileDestinations.LOCAL, path_in_storage)
				self._selectedFile = {
					"filename": path_in_storage,
					"filesize": filesize,
					"sd": sd,
					"estimatedPrintTime": None,
					"user": user
				}
			else:
				self._selectedFile = None
				self._stateMonitor.set_job_data(self._dict(file=self._dict(name=None,
				                                                           path=None,
				                                                           display=None,
				                                                           origin=None,
				                                                           size=None,
				                                                           date=None),
				                                           estimatedPrintTime=None,
				                                           averagePrintTime=None,
				                                           lastPrintTime=None,
				                                           filament=None,
				                                           user=None))
				return

			estimatedPrintTime = None
			lastPrintTime = None
			averagePrintTime = None
			date = None
			filament = None
			display_name = name_in_storage
			if path_on_disk:
				# Use a string for mtime because it could be float and the
				# javascript needs to exact match
				if not sd:
					date = int(os.stat(path_on_disk).st_mtime)

				try:
					fileData = self._fileManager.get_metadata(FileDestinations.SDCARD if sd else FileDestinations.LOCAL, path_on_disk)
				except Exception:
					self._logger.exception("Error generating fileData")
					fileData = None
				if fileData is not None:
					if "display" in fileData:
						display_name = fileData["display"]
					if "analysis" in fileData:
						if estimatedPrintTime is None and "estimatedPrintTime" in fileData["analysis"]:
							estimatedPrintTime = fileData["analysis"]["estimatedPrintTime"]
						if "filament" in fileData["analysis"]:
							filament = fileData["analysis"]["filament"]
					if "statistics" in fileData:
						printer_profile = self._printerProfileManager.get_current_or_default()["id"]
						if "averagePrintTime" in fileData["statistics"] and printer_profile in fileData["statistics"]["averagePrintTime"]:
							averagePrintTime = fileData["statistics"]["averagePrintTime"][printer_profile]
						if "lastPrintTime" in fileData["statistics"] and printer_profile in fileData["statistics"]["lastPrintTime"]:
							lastPrintTime = fileData["statistics"]["lastPrintTime"][printer_profile]

					if averagePrintTime is not None:
						self._selectedFile["estimatedPrintTime"] = averagePrintTime
						self._selectedFile["estimatedPrintTimeType"] = "average"
					elif estimatedPrintTime is not None:
						# TODO apply factor which first needs to be tracked!
						self._selectedFile["estimatedPrintTime"] = estimatedPrintTime
						self._selectedFile["estimatedPrintTimeType"] = "analysis"

			self._stateMonitor.set_job_data(self._dict(file=self._dict(name=name_in_storage,
			                                                           path=path_in_storage,
			                                                           display=display_name,
			                                                           origin=FileDestinations.SDCARD if sd else FileDestinations.LOCAL,
			                                                           size=filesize,
			                                                           date=date),
			                                           estimatedPrintTime=estimatedPrintTime,
			                                           averagePrintTime=averagePrintTime,
			                                           lastPrintTime=lastPrintTime,
			                                           filament=filament,
			                                           user=user))

	def _updateJobUser(self, user):
		with self._selectedFileMutex:
			if self._selectedFile is not None and self._selectedFile.get("user", None) != user:
				self._selectedFile["user"] = user

				job_data = self.get_current_job()
				self._stateMonitor.set_job_data(self._dict(file=job_data["file"],
				                                           estimatedPrintTime=job_data["estimatedPrintTime"],
				                                           averagePrintTime=job_data["averagePrintTime"],
				                                           lastPrintTime=job_data["lastPrintTime"],
				                                           filament=job_data["filament"],
				                                           user=user))

	def _sendInitialStateUpdate(self, callback):
		try:
			data = self._stateMonitor.get_current_data()
			data.update({
				"temps": list(self._temps),
				"logs": list(self._log),
				"messages": list(self._messages)
			})
			callback.on_printer_send_initial_data(data)
		except Exception:
			self._logger.exception(u"Error while pushing initial state update to callback {}".format(callback),
			                       extra=dict(callback=fqcn(callback)))

	def _getStateFlags(self):
		return self._dict(operational=self.is_operational(),
		                  printing=self.is_printing(),
		                  cancelling=self.is_cancelling(),
		                  pausing=self.is_pausing(),
		                  resuming=self.is_resuming(),
		                  finishing=self.is_finishing(),
		                  closedOrError=self.is_closed_or_error(),
		                  error=self.is_error(),
		                  paused=self.is_paused(),
		                  ready=self.is_ready(),
		                  sdReady=self.is_sd_ready())

	#~~ comm.MachineComPrintCallback implementation

	def on_comm_log(self, message):
		"""
		 Callback method for the comm object, called upon log output.
		"""
		self._addLog(to_unicode(message, "utf-8", errors="replace"))

	def on_comm_temperature_update(self, temp, bedTemp, chamberTemp):
		self._addTemperatureData(tools=copy.deepcopy(temp), bed=copy.deepcopy(bedTemp), chamber=copy.deepcopy(chamberTemp))

	def on_comm_position_update(self, position, reason=None):
		payload = dict(reason=reason)
		payload.update(position)
		eventManager().fire(Events.POSITION_UPDATE, payload)

	def on_comm_state_change(self, state):
		"""
		 Callback method for the comm object, called if the connection state changes.
		"""
		oldState = self._state

		state_string = None
		if self._comm is not None:
			state_string = self._comm.getStateString()

		if oldState in (comm.MachineCom.STATE_PRINTING,):
			# if we were still printing and went into an error state, mark the print as failed
			if state in (comm.MachineCom.STATE_CLOSED, comm.MachineCom.STATE_ERROR, comm.MachineCom.STATE_CLOSED_WITH_ERROR):
				with self._selectedFileMutex:
					if self._selectedFile is not None:
							payload = self._payload_for_print_job_event()
							if payload:
								payload["time"] = self._comm.getPrintTime()
								payload["reason"] = "error"
								payload["error"] = self._comm.getErrorString()

								def finalize():
									self._fileManager.log_print(payload["origin"],
									                            payload["path"],
									                            time.time(),
									                            payload["time"],
									                            False,
									                            self._printerProfileManager.get_current_or_default()["id"])
									eventManager().fire(Events.PRINT_FAILED, payload)

								thread = threading.Thread(target=finalize)
								thread.daemon = True
								thread.start()
			self._analysisQueue.resume() # printing done, put those cpu cycles to good use

		elif state == comm.MachineCom.STATE_PRINTING:
			if settings().get(["gcodeAnalysis", "runAt"]) == "idle":
				self._analysisQueue.pause() # only analyse files while idle

		if state == comm.MachineCom.STATE_CLOSED or state == comm.MachineCom.STATE_CLOSED_WITH_ERROR:
			if self._comm is not None:
				self._comm = None

			self._updateProgressData()
			self._setCurrentZ(None)
			self._setJobData(None, None, None)
			self._setOffsets(None)
			self._addTemperatureData()
			self._printerProfileManager.deselect()
			eventManager().fire(Events.DISCONNECTED)

		self._setState(state, state_string=state_string)

	def on_comm_message(self, message):
		"""
		 Callback method for the comm object, called upon message exchanges via serial.
		 Stores the message in the message buffer, truncates buffer to the last 300 lines.
		"""
		self._addMessage(to_unicode(message, "utf-8", errors="replace"))

	def on_comm_progress(self):
		"""
		 Callback method for the comm object, called upon any change in progress of the printjob.
		 Triggers storage of new values for printTime, printTimeLeft and the current progress.
		"""

		self._stateMonitor.trigger_progress_update()

	def on_comm_z_change(self, newZ):
		"""
		 Callback method for the comm object, called upon change of the z-layer.
		"""
		oldZ = self._currentZ
		if newZ != oldZ:
			# we have to react to all z-changes, even those that might "go backward" due to a slicer's retraction or
			# anti-backlash-routines. Event subscribes should individually take care to filter out "wrong" z-changes
			eventManager().fire(Events.Z_CHANGE, {"new": newZ, "old": oldZ})

		self._setCurrentZ(newZ)

	def on_comm_sd_state_change(self, sdReady):
		self._stateMonitor.set_state(self._dict(text=self.get_state_string(), flags=self._getStateFlags()))

	def on_comm_sd_files(self, files):
		eventManager().fire(Events.UPDATED_FILES, {"type": "printables"})
		self._sdFilelistAvailable.set()

	def on_comm_file_selected(self, full_path, size, sd, user=None):
		if full_path is not None:
			payload = self._payload_for_print_job_event(location=FileDestinations.SDCARD if sd else FileDestinations.LOCAL,
			                                            print_job_file=full_path,
			                                            print_job_user=user,
			                                            action_user=user)
			eventManager().fire(Events.FILE_SELECTED, payload)
			self._logger_job.info("Print job selected - origin: {}, path: {}, owner: {}, user: {}".format(payload.get("origin"),
			                                                                                              payload.get("path"),
			                                                                                              payload.get("owner"),
			                                                                                              payload.get("user")))
		else:
			eventManager().fire(Events.FILE_DESELECTED)
			self._logger_job.info("Print job deselected - user: {}".format(user if user else "n/a"))

		self._setJobData(full_path, size, sd, user=user)
		self._stateMonitor.set_state(self._dict(text=self.get_state_string(), flags=self._getStateFlags()))

		self._create_estimator()

		if self._printAfterSelect:
			self._printAfterSelect = False
			self.start_print(pos=self._posAfterSelect, user=user)

	def on_comm_print_job_started(self, suppress_script=False, user=None):
		self._updateJobUser(user) # the final job owner should always be whoever _started_ the job
		self._stateMonitor.trigger_progress_update()
		payload = self._payload_for_print_job_event(print_job_user=user,
		                                            action_user=user)
		if payload:
			eventManager().fire(Events.PRINT_STARTED, payload)
			self._logger_job.info("Print job started - origin: {}, path: {}, owner: {}, user: {}".format(payload.get("origin"),
			                                                                                             payload.get("path"),
			                                                                                             payload.get("owner"),
			                                                                                             payload.get("user")))

			if not suppress_script:
				self.script("beforePrintStarted",
				            context=dict(event=payload),
				            part_of_job=True,
				            must_be_set=False)

	def on_comm_print_job_done(self, suppress_script=False):
		self._fileManager.delete_recovery_data()

		payload = self._payload_for_print_job_event()
		if payload:
			payload["time"] = self._comm.getPrintTime()
			self._updateProgressData(completion=1.0,
			                         filepos=payload["size"],
			                         printTime=payload["time"],
			                         printTimeLeft=0)
			self._stateMonitor.set_state(self._dict(text=self.get_state_string(), flags=self._getStateFlags()))

			eventManager().fire(Events.PRINT_DONE, payload)
			self._logger_job.info("Print job done - origin: {}, path: {}, owner: {}".format(payload.get("origin"),
			                                                                                payload.get("path"),
			                                                                                payload.get("owner")))

			if not suppress_script:
				self.script("afterPrintDone",
				            context=dict(event=payload),
				            part_of_job=True,
				            must_be_set=False)

			def log_print():
				self._fileManager.log_print(payload["origin"],
				                            payload["path"],
				                            time.time(),
				                            payload["time"],
				                            True,
				                            self._printerProfileManager.get_current_or_default()["id"])

			thread = threading.Thread(target=log_print)
			thread.daemon = True
			thread.start()

		else:
			self._updateProgressData()
			self._stateMonitor.set_state(self._dict(text=self.get_state_string(), flags=self._getStateFlags()))


	def on_comm_print_job_cancelling(self, firmware_error=None, user=None):
		payload = self._payload_for_print_job_event(action_user=user)
		if payload:
			if firmware_error:
				payload["firmwareError"] = firmware_error
			eventManager().fire(Events.PRINT_CANCELLING, payload)

	def on_comm_print_job_cancelled(self, suppress_script=False, user=None):
		self._setCurrentZ(None)
		self._updateProgressData()

		payload = self._payload_for_print_job_event(position=self._comm.cancel_position.as_dict() if self._comm and self._comm.cancel_position else None,
		                                            action_user=user)
		if payload:
			payload["time"] = self._comm.getPrintTime()

			eventManager().fire(Events.PRINT_CANCELLED, payload)
			self._logger_job.info("Print job cancelled - origin: {}, path: {}, owner: {}, user: {}".format(payload.get("origin"),
			                                                                                               payload.get("path"),
			                                                                                               payload.get("owner"),
			                                                                                               payload.get("user")))

			if not suppress_script:
				self.script("afterPrintCancelled",
				            context=dict(event=payload),
				            part_of_job=True,
				            must_be_set=False)

			payload["reason"] = "cancelled"

			def finalize():
				self._fileManager.log_print(payload["origin"],
				                            payload["path"],
				                            time.time(),
				                            payload["time"],
				                            False,
				                            self._printerProfileManager.get_current_or_default()["id"])
				eventManager().fire(Events.PRINT_FAILED, payload)

			thread = threading.Thread(target=finalize)
			thread.daemon = True
			thread.start()

	def on_comm_print_job_paused(self, suppress_script=False, user=None):
		payload = self._payload_for_print_job_event(position=self._comm.pause_position.as_dict() if self._comm and self._comm.pause_position and not suppress_script else None,
		                                            action_user=user)
		if payload:
			eventManager().fire(Events.PRINT_PAUSED, payload)
			self._logger_job.info("Print job paused - origin: {}, path: {}, owner: {}, user: {}".format(payload.get("origin"),
			                                                                                            payload.get("path"),
			                                                                                            payload.get("owner"),
			                                                                                            payload.get("user")))
			if not suppress_script:
				self.script("afterPrintPaused",
				            context=dict(event=payload),
				            part_of_job=True,
				            must_be_set=False)

	def on_comm_print_job_resumed(self, suppress_script=False, user=None):
		payload = self._payload_for_print_job_event(action_user=user)
		if payload:
			eventManager().fire(Events.PRINT_RESUMED, payload)
			self._logger_job.info("Print job resumed - origin: {}, path: {}, owner: {}, user: {}".format(payload.get("origin"),
			                                                                                             payload.get("path"),
			                                                                                             payload.get("owner"),
			                                                                                             payload.get("user")))

			if not suppress_script:
				self.script("beforePrintResumed",
				            context=dict(event=payload),
				            part_of_job=True,
				            must_be_set=False)

	def on_comm_file_transfer_started(self, local_filename, remote_filename, filesize, user=None):
		eventManager().fire(Events.TRANSFER_STARTED, dict(local=local_filename,
		                                                  remote=remote_filename))

		self._sdStreaming = True

		self._setJobData(remote_filename, filesize, True, user=user)
		self._updateProgressData(completion=0.0, filepos=0, printTime=0)
		self._stateMonitor.set_state(self._dict(text=self.get_state_string(), flags=self._getStateFlags()))

	def on_comm_file_transfer_done(self, local_filename, remote_filename, elapsed, failed=False):
		self._sdStreaming = False

		payload = dict(local=local_filename,
		               remote=remote_filename,
		               time=elapsed)

		if failed:
			eventManager().fire(Events.TRANSFER_FAILED, payload)
			if callable(self._streamingFailedCallback):
				self._streamingFailedCallback(remote_filename, remote_filename, FileDestinations.SDCARD)
		else:
			eventManager().fire(Events.TRANSFER_DONE, payload)
			if callable(self._streamingFinishedCallback):
				self._streamingFinishedCallback(remote_filename, remote_filename, FileDestinations.SDCARD)

		self._setCurrentZ(None)
		self._setJobData(None, None, None)
		self._updateProgressData()
		self._stateMonitor.set_state(self._dict(text=self.get_state_string(), flags=self._getStateFlags()))

	def on_comm_file_transfer_failed(self, local_filename, remote_filename, elapsed):
		self.on_comm_file_transfer_done(local_filename, remote_filename, elapsed, failed=True)

	def on_comm_force_disconnect(self):
		self.disconnect()

	def on_comm_record_fileposition(self, origin, name, pos):
		try:
			self._fileManager.save_recovery_data(origin, name, pos)
		except NoSuchStorage:
			pass
		except Exception:
			self._logger.exception("Error while trying to persist print recovery data")

	def _payload_for_print_job_event(self, location=None, print_job_file=None, print_job_size=None, print_job_user=None,
	                                 position=None, action_user=None):
		if print_job_file is None:
			with self._selectedFileMutex:
				selected_file = self._selectedFile
				if not selected_file:
					return dict()

				print_job_file = selected_file.get("filename", None)
				print_job_size = selected_file.get("filesize", None)
				print_job_user = selected_file.get("user", None)
				location = FileDestinations.SDCARD if selected_file.get("sd", False) else FileDestinations.LOCAL

		if not print_job_file or not location:
			return dict()

		if location == FileDestinations.SDCARD:
			full_path = print_job_file
			if full_path.startswith("/"):
				full_path = full_path[1:]
			name = path = full_path
			origin = FileDestinations.SDCARD

		else:
			full_path = self._fileManager.path_on_disk(FileDestinations.LOCAL, print_job_file)
			path = self._fileManager.path_in_storage(FileDestinations.LOCAL, print_job_file)
			_, name = self._fileManager.split_path(FileDestinations.LOCAL, path)
			origin = FileDestinations.LOCAL

		result= dict(name=name,
		             path=path,
		             origin=origin,
		             size=print_job_size)

		if position is not None:
			result["position"] = position

		if print_job_user is not None:
			result["owner"] = print_job_user

		if action_user is not None:
			result["user"] = action_user

		return result


class StateMonitor(object):
	def __init__(self, interval=0.5, on_update=None, on_add_temperature=None, on_add_log=None, on_add_message=None, on_get_progress=None):
		self._interval = interval
		self._update_callback = on_update
		self._on_add_temperature = on_add_temperature
		self._on_add_log = on_add_log
		self._on_add_message = on_add_message
		self._on_get_progress = on_get_progress

		self._state = None
		self._job_data = None
		self._current_z = None
		self._offsets = dict()
		self._progress = None

		self._progress_dirty = False

		self._change_event = threading.Event()
		self._state_lock = threading.Lock()
		self._progress_lock = threading.Lock()

		self._last_update = monotonic_time()
		self._worker = threading.Thread(target=self._work)
		self._worker.daemon = True
		self._worker.start()

	def _get_current_progress(self):
		if callable(self._on_get_progress):
			return self._on_get_progress()
		return self._progress

	def reset(self, state=None, job_data=None, progress=None, current_z=None, offsets=None):
		self.set_state(state)
		self.set_job_data(job_data)
		self.set_progress(progress)
		self.set_current_z(current_z)
		self.set_temp_offsets(offsets)

	def add_temperature(self, temperature):
		self._on_add_temperature(temperature)
		self._change_event.set()

	def add_log(self, log):
		self._on_add_log(log)
		self._change_event.set()

	def add_message(self, message):
		self._on_add_message(message)
		self._change_event.set()

	def set_current_z(self, current_z):
		self._current_z = current_z
		self._change_event.set()

	def set_state(self, state):
		with self._state_lock:
			self._state = state
			self._change_event.set()

	def set_job_data(self, job_data):
		self._job_data = job_data
		self._change_event.set()

	def trigger_progress_update(self):
		with self._progress_lock:
			self._progress_dirty = True
			self._change_event.set()

	def set_progress(self, progress):
		with self._progress_lock:
			self._progress_dirty = False
			self._progress = progress
			self._change_event.set()

	def set_temp_offsets(self, offsets):
		if offsets is None:
			offsets = dict()
		self._offsets = offsets
		self._change_event.set()

	def _work(self):
		try:
			while True:
				self._change_event.wait()

				now = monotonic_time()
				delta = now - self._last_update
				additional_wait_time = self._interval - delta
				if additional_wait_time > 0:
					time.sleep(additional_wait_time)

				with self._state_lock:
					data = self.get_current_data()
					self._update_callback(data)
					self._last_update = monotonic_time()
					self._change_event.clear()
		except Exception:
			logging.getLogger(__name__).exception("Looks like something crashed inside the state update worker. Please report this on the OctoPrint issue tracker (make sure to include logs!)")

	def get_current_data(self):
		with self._progress_lock:
			if self._progress_dirty:
				self._progress = self._get_current_progress()
				self._progress_dirty = False

		return {
			"state": self._state,
			"job": self._job_data,
			"currentZ": self._current_z,
			"progress": self._progress,
			"offsets": self._offsets
		}


class TemperatureHistory(InvariantContainer):
	def __init__(self, cutoff=30 * 60):

		def temperature_invariant(data):
			data.sort(key=lambda x: x["time"])
			now = int(time.time())
			return [item for item in data if item["time"] >= now - cutoff]

		InvariantContainer.__init__(self, guarantee_invariant=temperature_invariant)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import os
import threading
import logging

from flask import request, jsonify, url_for, make_response

import octoprint.timelapse
import octoprint.util as util
from octoprint.settings import settings, valid_boolean_trues

from octoprint.server import admin_permission, printer
from octoprint.server.util.flask import redirect_to_tornado, no_firstrun_access, get_json_command_from_request, with_revalidation_checking
from octoprint.server.api import api

from octoprint.server import NO_CONTENT
from octoprint.access.permissions import Permissions

_DATA_FORMAT_VERSION = "v2"

#~~ timelapse handling

_timelapse_cache_finished = []
_timelapse_cache_finished_lastmodified = None
_timelapse_cache_unrendered = []
_timelapse_cache_unrendered_lastmodified = None
_timelapse_cache_mutex = threading.RLock()

def _config_for_timelapse(timelapse):
	if timelapse is not None and isinstance(timelapse, octoprint.timelapse.ZTimelapse):
		return dict(type="zchange",
		            postRoll=timelapse.post_roll,
		            fps=timelapse.fps,
		            retractionZHop=timelapse.retraction_zhop,
		            minDelay=timelapse.min_delay)
	elif timelapse is not None and isinstance(timelapse, octoprint.timelapse.TimedTimelapse):
		return dict(type="timed",
		            postRoll=timelapse.post_roll,
		            fps=timelapse.fps,
		            interval=timelapse.interval)
	else:
		return dict(type="off")

def _lastmodified(unrendered):
	lm_finished = octoprint.timelapse.last_modified_finished()
	if unrendered:
		lm_unrendered = octoprint.timelapse.last_modified_unrendered()

		if lm_finished is None or lm_unrendered is None:
			return None
		return max(lm_finished, lm_unrendered)
	return lm_finished

def _etag(unrendered, lm=None):
	if lm is None:
		lm = _lastmodified(unrendered)

	timelapse = octoprint.timelapse.current
	config = _config_for_timelapse(timelapse)

	import hashlib
	hash = hashlib.sha1()
	def hash_update(value):
		value = value.encode('utf-8')
		hash.update(value)
	hash_update(str(lm))
	hash_update(repr(config))
	hash_update(repr(_DATA_FORMAT_VERSION))

	return hash.hexdigest()

@api.route("/timelapse", methods=["GET"])
@with_revalidation_checking(etag_factory=lambda lm=None: _etag(request.values.get("unrendered", "false") in valid_boolean_trues, lm=lm),
                            lastmodified_factory=lambda: _lastmodified(request.values.get("unrendered", "false") in valid_boolean_trues),
                            unless=lambda: request.values.get("force", "false") in valid_boolean_trues)
@no_firstrun_access
@Permissions.TIMELAPSE_LIST.require(403)
def getTimelapseData():
	timelapse = octoprint.timelapse.current
	config = _config_for_timelapse(timelapse)

	force = request.values.get("force", "false") in valid_boolean_trues
	unrendered = request.values.get("unrendered", "false") in valid_boolean_trues

	global _timelapse_cache_finished_lastmodified, _timelapse_cache_finished, _timelapse_cache_unrendered_lastmodified, _timelapse_cache_unrendered
	with _timelapse_cache_mutex:
		current_lastmodified_finished = octoprint.timelapse.last_modified_finished()
		current_lastmodified_unrendered = octoprint.timelapse.last_modified_unrendered()

		if not force and _timelapse_cache_finished_lastmodified == current_lastmodified_finished:
			files = _timelapse_cache_finished
		else:
			files = octoprint.timelapse.get_finished_timelapses()
			_timelapse_cache_finished = files
			_timelapse_cache_finished_lastmodified = current_lastmodified_finished

		unrendered_files = []
		if unrendered:
			if not force and _timelapse_cache_unrendered_lastmodified == current_lastmodified_unrendered:
				unrendered_files = _timelapse_cache_unrendered
			else:
				unrendered_files = octoprint.timelapse.get_unrendered_timelapses()
				_timelapse_cache_unrendered = unrendered_files
				_timelapse_cache_unrendered_lastmodified = current_lastmodified_unrendered

	finished_list = []
	for f in files:
		output = dict(f)
		output["url"] = url_for("index") + "downloads/timelapse/" + f["name"]
		finished_list.append(output)

	result = dict(config=config,
	              enabled=settings().getBoolean(["webcam", "timelapseEnabled"]),
	              files=finished_list)

	if unrendered:
		result.update(unrendered=unrendered_files)

	return jsonify(result)


@api.route("/timelapse/<filename>", methods=["GET"])
@no_firstrun_access
@Permissions.TIMELAPSE_DOWNLOAD.require(403)
def downloadTimelapse(filename):
	return redirect_to_tornado(request, url_for("index") + "downloads/timelapse/" + filename)


@api.route("/timelapse/<filename>", methods=["DELETE"])
@no_firstrun_access
@Permissions.TIMELAPSE_DELETE.require(403)
def deleteTimelapse(filename):
	timelapse_folder = settings().getBaseFolder("timelapse")
	full_path = os.path.realpath(os.path.join(timelapse_folder, filename))
	if octoprint.timelapse.valid_timelapse(full_path) \
			and full_path.startswith(timelapse_folder) \
			and os.path.exists(full_path) \
			and not util.is_hidden_path(full_path):
		try:
			os.remove(full_path)
		except Exception as ex:
			logging.getLogger(__file__).exception("Error deleting timelapse file {}".format(full_path))
			return make_response("Unexpected error: {}".format(ex), 500)

	return getTimelapseData()


@api.route("/timelapse/unrendered/<name>", methods=["DELETE"])
@no_firstrun_access
@Permissions.TIMELAPSE_DELETE.require(403)
def deleteUnrenderedTimelapse(name):
	octoprint.timelapse.delete_unrendered_timelapse(name)
	return NO_CONTENT


@api.route("/timelapse/unrendered/<name>", methods=["POST"])
@no_firstrun_access
@Permissions.TIMELAPSE_ADMIN.require(403)
def processUnrenderedTimelapseCommand(name):
	# valid file commands, dict mapping command name to mandatory parameters
	valid_commands = {
		"render": []
	}

	command, data, response = get_json_command_from_request(request, valid_commands)
	if response is not None:
		return response

	if command == "render":
		if printer.is_printing() or printer.is_paused():
			return make_response("Printer is currently printing, cannot render timelapse", 409)
		octoprint.timelapse.render_unrendered_timelapse(name)

	return NO_CONTENT


@api.route("/timelapse", methods=["POST"])
@no_firstrun_access
@Permissions.TIMELAPSE_ADMIN.require(403)
def setTimelapseConfig():
	data = request.get_json(silent=True)
	if data is None:
		data = request.values

	if "type" in data:
		config = {
			"type": data["type"],
			"postRoll": 0,
			"fps": 25,
			"options": {}
		}

		if "postRoll" in data:
			try:
				postRoll = int(data["postRoll"])
			except ValueError:
				return make_response("Invalid value for postRoll: %r" % data["postRoll"], 400)
			else:
				if postRoll >= 0:
					config["postRoll"] = postRoll
				else:
					return make_response("Invalid value for postRoll: %d" % postRoll, 400)

		if "fps" in data:
			try:
				fps = int(data["fps"])
			except ValueError:
				return make_response("Invalid value for fps: %r" % data["fps"], 400)
			else:
				if fps > 0:
					config["fps"] = fps
				else:
					return make_response("Invalid value for fps: %d" % fps, 400)

		if "interval" in data:
			try:
				interval = int(data["interval"])
			except ValueError:
				return make_response("Invalid value for interval: %r" % data["interval"], 400)
			else:
				if interval > 0:
					config["options"]["interval"] = interval
				else:
					return make_response("Invalid value for interval: %d" % interval, 400)

		if "retractionZHop" in data:
			try:
				retractionZHop = float(data["retractionZHop"])
			except ValueError:
				return make_response("Invalid value for retraction Z-Hop: %r" % data["retractionZHop"], 400)
			else:
				if retractionZHop >= 0:
					config["options"]["retractionZHop"] = retractionZHop
				else:
					return make_response("Invalid value for retraction Z-Hop: %f" % retractionZHop, 400)

		if "minDelay" in data:
			try:
				minDelay = float(data["minDelay"])
			except ValueError:
				return make_response("Invalid value for minimum delay: %r" % data["minDelay"], 400)
			else:
				if minDelay > 0:
					config["options"]["minDelay"] = minDelay
				else:
					return make_response("Invalid value for minimum delay: %f" % minDelay, 400)

		if admin_permission.can() and "save" in data and data["save"] in valid_boolean_trues:
			octoprint.timelapse.configure_timelapse(config, True)
		else:
			octoprint.timelapse.configure_timelapse(config)

	return getTimelapseData()




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


import errno
import subprocess
import sys
import logging

from ..exceptions import ConfigurationInvalid

def _get_git_executables():
	GITS = ["git"]
	if sys.platform == "win32":
		GITS = ["git.cmd", "git.exe"]
	return GITS


def _git(args, cwd, hide_stderr=False):
	commands = _get_git_executables()

	for c in commands:
		try:
			p = subprocess.Popen([c] + args, cwd=cwd, stdout=subprocess.PIPE,
			                     stderr=(subprocess.PIPE if hide_stderr
			                             else None))
			break
		except EnvironmentError:
			e = sys.exc_info()[1]
			if e.errno == errno.ENOENT:
				continue
			return None, None
	else:
		return None, None

	stdout = p.communicate()[0].strip()
	if sys.version >= '3':
		stdout = stdout.decode()

	if p.returncode != 0:
		return p.returncode, None

	return p.returncode, stdout


def get_latest(target, check, online=True):
	checkout_folder = check.get("checkout_folder")
	if checkout_folder is None:
		raise ConfigurationInvalid("Update configuration for {} of type git_commit needs checkout_folder set and not None".format(target))

	returncode, local_commit = _git(["rev-parse", "@{0}"], checkout_folder)
	if returncode != 0:
		return None, True

	information = dict(
		local=dict(name="Commit %s" % local_commit, value=local_commit),
		remote=dict(name="?", value="?"),
		needs_online=not check.get("offline", False)
	)
	if not online and information["needs_online"]:
		return information, True

	returncode, _ = _git(["fetch"], checkout_folder)
	if returncode != 0:
		return information, True

	returncode, remote_commit = _git(["rev-parse", "@{u}"], checkout_folder)
	if returncode != 0:
		return information, True

	returncode, base = _git(["merge-base", "@{0}", "@{u}"], checkout_folder)
	if returncode != 0:
		return information, True

	if local_commit == remote_commit or remote_commit == base:
		information["remote"] = dict(name="Commit %s" % local_commit, value=local_commit)
		is_current = True
	else:
		information["remote"] = dict(name="Commit %s" % remote_commit, value=remote_commit)
		is_current = local_commit == remote_commit

	logger = logging.getLogger("octoprint.plugins.softwareupdate.version_checks.git_commit")
	logger.debug("Target: %s, local: %s, remote: %s" % (target, local_commit, remote_commit))

	return information, is_current



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


import click
click.disable_unicode_literals_warning = True

from past.builtins import basestring

class OctoPrintDevelCommands(click.MultiCommand):
	"""
	Custom `click.MultiCommand <http://click.pocoo.org/5/api/#click.MultiCommand>`_
	implementation that provides commands relevant for (plugin) development
	based on availability of development dependencies.
	"""

	sep = ":"
	groups = ("plugin",)

	def __init__(self, *args, **kwargs):
		click.MultiCommand.__init__(self, *args, **kwargs)

		from octoprint.util.commandline import CommandlineCaller
		from functools import partial

		def log_util(f):
			def log(*lines):
					for line in lines:
						f(line)
			return log

		self.command_caller = CommandlineCaller()
		self.command_caller.on_log_call = log_util(lambda x: click.echo(">> {}".format(x)))
		self.command_caller.on_log_stdout = log_util(click.echo)
		self.command_caller.on_log_stderr = log_util(partial(click.echo, err=True))

	def _get_prefix_methods(self, method_prefix):
		for name in [x for x in dir(self) if x.startswith(method_prefix)]:
			method = getattr(self, name)
			yield method

	def _get_commands_from_prefix_methods(self, method_prefix):
		for method in self._get_prefix_methods(method_prefix):
			result = method()
			if result is not None and isinstance(result, click.Command):
				yield result

	def _get_commands(self):
		result = dict()
		for group in self.groups:
			for command in self._get_commands_from_prefix_methods("{}_".format(group)):
				result[group + self.sep + command.name] = command
		return result

	def list_commands(self, ctx):
		result = [name for name in self._get_commands()]
		result.sort()
		return result

	def get_command(self, ctx, cmd_name):
		commands = self._get_commands()
		return commands.get(cmd_name, None)

	def plugin_new(self):
		try:
			import cookiecutter.main
		except ImportError:
			return None

		try:
			# we depend on Cookiecutter >= 1.4
			from cookiecutter.prompt import StrictEnvironment
		except ImportError:
			return None

		import contextlib

		@contextlib.contextmanager
		def custom_cookiecutter_config(config):
			"""
			Allows overriding cookiecutter's user config with a custom dict
			with fallback to the original data.
			"""
			from octoprint.util import fallback_dict

			original_get_user_config = cookiecutter.main.get_user_config
			try:
				def f(*args, **kwargs):
					original_config = original_get_user_config(*args, **kwargs)
					return fallback_dict(config, original_config)
				cookiecutter.main.get_user_config = f
				yield
			finally:
				cookiecutter.main.get_user_config = original_get_user_config

		@contextlib.contextmanager
		def custom_cookiecutter_prompt(options):
			"""
			Custom cookiecutter prompter for the template config.

			If a setting is available in the provided options (read from the CLI)
			that will be used, otherwise the user will be prompted for a value
			via click.
			"""
			original_prompt_for_config = cookiecutter.main.prompt_for_config

			def custom_prompt_for_config(context, no_input=False):
				cookiecutter_dict = dict()

				env = StrictEnvironment()

				for key, raw in context['cookiecutter'].items():
					if key in options:
						val = options[key]
					else:
						if not isinstance(raw, basestring):
							raw = str(raw)
						val = env.from_string(raw).render(cookiecutter=cookiecutter_dict)

						if not no_input:
							val = click.prompt(key, default=val)

					cookiecutter_dict[key] = val
				return cookiecutter_dict

			try:
				cookiecutter.main.prompt_for_config = custom_prompt_for_config
				yield
			finally:
				cookiecutter.main.prompt_for_config = original_prompt_for_config

		@click.command("new")
		@click.option("--name", "-n", help="The name of the plugin")
		@click.option("--package", "-p", help="The plugin package")
		@click.option("--author", "-a", help="The plugin author's name")
		@click.option("--email", "-e", help="The plugin author's mail address")
		@click.option("--license", "-l", help="The plugin's license")
		@click.option("--description", "-d", help="The plugin's description")
		@click.option("--homepage", help="The plugin's homepage URL")
		@click.option("--source", "-s", help="The URL to the plugin's source")
		@click.option("--installurl", "-i", help="The plugin's install URL")
		@click.argument("identifier", required=False)
		def command(name, package, author, email, description, license, homepage, source, installurl, identifier):
			"""Creates a new plugin based on the OctoPrint Plugin cookiecutter template."""
			from octoprint.util import tempdir

			# deleting a git checkout folder might run into access errors due
			# to write-protected sub folders, so we use a custom onerror handler
			# that tries to fix such permissions
			def onerror(func, path, exc_info):
				"""Originally from http://stackoverflow.com/a/2656405/2028598"""
				import stat
				import os

				if not os.access(path, os.W_OK):
					os.chmod(path, stat.S_IWUSR)
					func(path)
				else:
					raise

			with tempdir(onerror=onerror) as path:
				custom = dict(cookiecutters_dir=path)
				with custom_cookiecutter_config(custom):
					raw_options = dict(
						plugin_identifier=identifier,
						plugin_package=package,
						plugin_name=name,
						full_name=author,
						email=email,
						plugin_description=description,
						plugin_license=license,
						plugin_homepage=homepage,
						plugin_source=source,
						plugin_installurl=installurl
					)
					options = dict((k, v) for k, v in raw_options.items() if v is not None)

					with custom_cookiecutter_prompt(options):
						cookiecutter.main.cookiecutter("gh:OctoPrint/cookiecutter-octoprint-plugin")

		return command

	def plugin_install(self):
		@click.command("install")
		@click.option("--path", help="Path of the local plugin development folder to install")
		def command(path):
			"""
			Installs the local plugin in development mode.

			Note: This can NOT be used to install plugins from remote locations
			such as the plugin repository! It is strictly for local development
			of plugins, to ensure the plugin is installed (editable) into the
			same python environment that OctoPrint is installed under.
			"""

			import os
			import sys

			if not path:
				path = os.getcwd()

			# check if this really looks like a plugin
			if not os.path.isfile(os.path.join(path, "setup.py")):
				click.echo("This doesn't look like an OctoPrint plugin folder")
				sys.exit(1)

			self.command_caller.call([sys.executable, "-m", "pip", "install", "-e", "."], cwd=path)

		return command

	def plugin_uninstall(self):
		@click.command("uninstall")
		@click.argument("name")
		def command(name):
			"""Uninstalls the plugin with the given name."""
			import sys

			lower_name = name.lower()
			if not lower_name.startswith("octoprint_") and not lower_name.startswith("octoprint-"):
				click.echo("This doesn't look like an OctoPrint plugin name")
				sys.exit(1)

			call = [sys.executable, "-m", "pip", "uninstall", "--yes", name]
			self.command_caller.call(call)

		return command

@click.group()
def dev_commands():
	pass

@dev_commands.group(name="dev", cls=OctoPrintDevelCommands)
def dev():
	"""Additional commands for development tasks."""
	pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


import click
click.disable_unicode_literals_warning = True
import logging

from octoprint import init_settings, FatalStartupError
from octoprint.cli import standard_options, bulk_options, get_ctx_obj_option

import yaml
import json
import pprint

def _to_settings_path(path):
	if not isinstance(path, (list, tuple)):
		path = list(filter(lambda x: x, map(lambda x: x.strip(), path.split("."))))
	return path

def _set_helper(settings, path, value, data_type=None):
	path = _to_settings_path(path)

	method = settings.set
	if data_type is not None:
		name = None
		if data_type == bool:
			name = "setBoolean"
		elif data_type == float:
			name = "setFloat"
		elif data_type == int:
			name = "setInt"

		if name is not None:
			method = getattr(settings, name)

	method(path, value)
	settings.save()

#~~ "octoprint config" commands

@click.group()
def config_commands():
	pass

@config_commands.group(name="config")
@click.pass_context
def config(ctx):
	"""Basic config manipulation."""
	logging.basicConfig(level=logging.DEBUG if get_ctx_obj_option(ctx, "verbosity", 0) > 0 else logging.WARN)
	try:
		ctx.obj.settings = init_settings(get_ctx_obj_option(ctx, "basedir", None), get_ctx_obj_option(ctx, "configfile", None))
	except FatalStartupError as e:
		click.echo(e.message, err=True)
		click.echo("There was a fatal error initializing the settings manager.", err=True)
		ctx.exit(-1)


@config.command(name="set")
@standard_options(hidden=True)
@click.argument("path", type=click.STRING)
@click.argument("value", type=click.STRING)
@click.option("--bool", "as_bool", is_flag=True,
              help="Interpret value as bool")
@click.option("--float", "as_float", is_flag=True,
              help="Interpret value as float")
@click.option("--int", "as_int", is_flag=True,
              help="Interpret value as int")
@click.option("--json", "as_json", is_flag=True,
              help="Parse value from json")
@click.pass_context
def set_command(ctx, path, value, as_bool, as_float, as_int, as_json):
	"""Sets a config path to the provided value."""
	if as_json:
		try:
			value = json.loads(value)
		except Exception as e:
			click.echo(e.message, err=True)
			ctx.exit(-1)

	data_type = None
	if as_bool:
		data_type = bool
	elif as_float:
		data_type = float
	elif as_int:
		data_type = int

	_set_helper(ctx.obj.settings, path, value, data_type=data_type)


@config.command(name="remove")
@standard_options(hidden=True)
@click.argument("path", type=click.STRING)
@click.pass_context
def remove_command(ctx, path):
	"""Removes a config path."""
	_set_helper(ctx.obj.settings, path, None)


@config.command(name="append_value")
@standard_options(hidden=True)
@click.argument("path", type=click.STRING)
@click.argument("value", type=click.STRING)
@click.option("--json", "as_json", is_flag=True)
@click.pass_context
def append_value_command(ctx, path, value, as_json=False):
	"""Appends value to list behind config path."""
	path = _to_settings_path(path)

	if as_json:
		try:
			value = json.loads(value)
		except Exception as e:
			click.echo(e.message, err=True)
			ctx.exit(-1)

	current = ctx.obj.settings.get(path)
	if current is None:
		current = []
	if not isinstance(current, list):
		click.echo("Cannot append to non-list value at given path", err=True)
		ctx.exit(-1)

	current.append(value)
	_set_helper(ctx.obj.settings, path, current)


@config.command(name="insert_value")
@standard_options(hidden=True)
@click.argument("path", type=click.STRING)
@click.argument("index", type=click.INT)
@click.argument("value", type=click.STRING)
@click.option("--json", "as_json", is_flag=True)
@click.pass_context
def insert_value_command(ctx, path, index, value, as_json=False):
	"""Inserts value at index of list behind config key."""
	path = _to_settings_path(path)

	if as_json:
		try:
			value = json.loads(value)
		except Exception as e:
			click.echo(e.message, err=True)
			ctx.exit(-1)

	current = ctx.obj.settings.get(path)
	if current is None:
		current = []
	if not isinstance(current, list):
		click.echo("Cannot insert into non-list value at given path", err=True)
		ctx.exit(-1)

	current.insert(index, value)
	_set_helper(ctx.obj.settings, path, current)


@config.command(name="remove_value")
@standard_options(hidden=True)
@click.argument("path", type=click.STRING)
@click.argument("value", type=click.STRING)
@click.option("--json", "as_json", is_flag=True)
@click.pass_context
def remove_value_command(ctx, path, value, as_json=False):
	"""Removes value from list at config path."""
	path = _to_settings_path(path)

	if as_json:
		try:
			value = json.loads(value)
		except Exception as e:
			click.echo(e.message, err=True)
			ctx.exit(-1)

	current = ctx.obj.settings.get(path)
	if current is None:
		current = []
	if not isinstance(current, list):
		click.echo("Cannot remove value from non-list value at given path", err=True)
		ctx.exit(-1)

	if not value in current:
		click.echo("Value is not contained in list at given path")
		ctx.exit()

	current.remove(value)
	_set_helper(ctx.obj.settings, path, current)


@config.command(name="get")
@click.argument("path", type=click.STRING)
@click.option("--json", "as_json", is_flag=True,
              help="Output value formatted as JSON")
@click.option("--yaml", "as_yaml", is_flag=True,
              help="Output value formatted as YAML")
@click.option("--raw", "as_raw", is_flag=True,
              help="Output value as raw string representation")
@standard_options(hidden=True)
@click.pass_context
def get_command(ctx, path, as_json=False, as_yaml=False, as_raw=False):
	"""Retrieves value from config path."""
	path = _to_settings_path(path)
	value = ctx.obj.settings.get(path, merged=True)

	if as_json:
		output = json.dumps(value)
	elif as_yaml:
		output = yaml.safe_dump(value, default_flow_style=False, indent=4, allow_unicode=True)
	elif as_raw:
		output = value
	else:
		output = pprint.pformat(value)

	click.echo(output)


@config.command(name="effective")
@click.option("--json", "as_json", is_flag=True,
              help="Output value formatted as JSON")
@click.option("--yaml", "as_yaml", is_flag=True,
              help="Output value formatted as YAML")
@click.option("--raw", "as_raw", is_flag=True,
              help="Output value as raw string representation")
@standard_options(hidden=True)
@click.pass_context
def effective_command(ctx, as_json=False, as_yaml=False, as_raw=False):
		"""Retrieves the full effective config."""
		value = ctx.obj.settings.effective

		if as_json:
				output = json.dumps(value)
		elif as_yaml:
				output = yaml.safe_dump(value, default_flow_style=False, indent=4, allow_unicode=True)
		elif as_raw:
				output = value
		else:
				output = pprint.pformat(value)

		click.echo(output)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

from flask import request, jsonify, make_response, url_for
from werkzeug.exceptions import BadRequest

from octoprint.server import slicingManager
from octoprint.server.util.flask import no_firstrun_access, with_revalidation_checking
from octoprint.server.api import api, NO_CONTENT
from octoprint.access.permissions import Permissions

from octoprint.settings import settings as s, valid_boolean_trues

from octoprint.slicing import UnknownSlicer, SlicerNotConfigured, UnknownProfile, CouldNotDeleteProfile


_DATA_FORMAT_VERSION = "v2"


def _lastmodified(configured):
	if configured:
		slicers = slicingManager.configured_slicers
	else:
		slicers = slicingManager.registered_slicers

	lms = [0]
	for slicer in slicers:
		lms.append(slicingManager.profiles_last_modified(slicer))

	return max(lms)


def _etag(configured, lm=None):
	if lm is None:
		lm = _lastmodified(configured)

	import hashlib
	hash = hashlib.sha1()
	def hash_update(value):
		value = value.encode('utf-8')
		hash.update(value)
	hash_update(str(lm))

	if configured:
		slicers = slicingManager.configured_slicers
	else:
		slicers = slicingManager.registered_slicers

	default_slicer = s().get(["slicing", "defaultSlicer"])

	for slicer in sorted(slicers):
		slicer_impl = slicingManager.get_slicer(slicer, require_configured=False)
		hash_update(slicer)
		hash_update(str(slicer_impl.is_slicer_configured()))
		hash_update(str(slicer == default_slicer))

	hash_update(_DATA_FORMAT_VERSION) # increment version if we change the API format

	return hash.hexdigest()


@api.route("/slicing", methods=["GET"])
@with_revalidation_checking(etag_factory=lambda lm=None: _etag(request.values.get("configured", "false") in valid_boolean_trues, lm=lm),
                            lastmodified_factory=lambda: _lastmodified(request.values.get("configured", "false") in valid_boolean_trues),
                            unless=lambda: request.values.get("force", "false") in valid_boolean_trues)
@Permissions.SLICE.require(403)
def slicingListAll():
	from octoprint.filemanager import get_extensions

	default_slicer = s().get(["slicing", "defaultSlicer"])

	if "configured" in request.values and request.values["configured"] in valid_boolean_trues:
		slicers = slicingManager.configured_slicers
	else:
		slicers = slicingManager.registered_slicers

	result = dict()
	for slicer in slicers:
		try:
			slicer_impl = slicingManager.get_slicer(slicer, require_configured=False)

			extensions = set()
			for source_file_type in slicer_impl.get_slicer_properties().get("source_file_types", ["model"]):
				extensions = extensions.union(get_extensions(source_file_type))

			result[slicer] = dict(
				key=slicer,
				displayName=slicer_impl.get_slicer_properties().get("name", "n/a"),
				sameDevice=slicer_impl.get_slicer_properties().get("same_device", True),
				default=default_slicer == slicer,
				configured=slicer_impl.is_slicer_configured(),
				profiles=_getSlicingProfilesData(slicer),
				extensions=dict(
					source=list(extensions),
					destination=slicer_impl.get_slicer_properties().get("destination_extensions", ["gco", "gcode", "g"])
				)
			)
		except (UnknownSlicer, SlicerNotConfigured):
			# this should never happen
			pass

	return jsonify(result)

@api.route("/slicing/<string:slicer>/profiles", methods=["GET"])
@no_firstrun_access
@Permissions.SLICE.require(403)
def slicingListSlicerProfiles(slicer):
	configured = False
	if "configured" in request.values and request.values["configured"] in valid_boolean_trues:
		configured = True

	try:
		return jsonify(_getSlicingProfilesData(slicer, require_configured=configured))
	except (UnknownSlicer, SlicerNotConfigured):
		return make_response("Unknown slicer {slicer}".format(**locals()), 404)

@api.route("/slicing/<string:slicer>/profiles/<string:name>", methods=["GET"])
@no_firstrun_access
@Permissions.SLICE.require(403)
def slicingGetSlicerProfile(slicer, name):
	try:
		profile = slicingManager.load_profile(slicer, name, require_configured=False)
	except UnknownSlicer:
		return make_response("Unknown slicer {slicer}".format(**locals()), 404)
	except UnknownProfile:
		return make_response("Profile not found", 404)

	result = _getSlicingProfileData(slicer, name, profile)
	result["data"] = profile.data
	return jsonify(result)

@api.route("/slicing/<string:slicer>/profiles/<string:name>", methods=["PUT"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def slicingAddSlicerProfile(slicer, name):
	if not "application/json" in request.headers["Content-Type"]:
		return make_response("Expected content-type JSON", 400)

	try:
		json_data = request.get_json()
	except BadRequest:
		return make_response("Malformed JSON body in request", 400)

	if json_data is None:
		return make_response("Malformed JSON body in request", 400)

	data = dict()
	display_name = None
	description = None
	if "data" in json_data:
		data = json_data["data"]
	if "displayName" in json_data:
		display_name = json_data["displayName"]
	if "description" in json_data:
		description = json_data["description"]

	try:
		profile = slicingManager.save_profile(slicer, name, data,
		                                      allow_overwrite=True, display_name=display_name, description=description)
	except UnknownSlicer:
		return make_response("Unknown slicer {slicer}".format(**locals()), 404)

	result = _getSlicingProfileData(slicer, name, profile)
	r = make_response(jsonify(result), 201)
	r.headers["Location"] = result["resource"]
	return r

@api.route("/slicing/<string:slicer>/profiles/<string:name>", methods=["PATCH"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def slicingPatchSlicerProfile(slicer, name):
	if not "application/json" in request.headers["Content-Type"]:
		return make_response("Expected content-type JSON", 400)

	try:
		profile = slicingManager.load_profile(slicer, name, require_configured=False)
	except UnknownSlicer:
		return make_response("Unknown slicer {slicer}".format(**locals()), 404)
	except UnknownProfile:
		return make_response("Profile {name} for slicer {slicer} not found".format(**locals()), 404)

	try:
		json_data = request.get_json()
	except BadRequest:
		return make_response("Malformed JSON body in request", 400)

	if json_data is None:
		return make_response("Malformed JSON body in request", 400)

	data = dict()
	display_name = None
	description = None
	if "data" in json_data:
		data = json_data["data"]
	if "displayName" in json_data:
		display_name = json_data["displayName"]
	if "description" in json_data:
		description = json_data["description"]

	saved_profile = slicingManager.save_profile(slicer, name, profile,
	                                            allow_overwrite=True,
	                                            overrides=data,
	                                            display_name=display_name,
	                                            description=description)

	from octoprint.server.api import valid_boolean_trues
	if "default" in json_data and json_data["default"] in valid_boolean_trues:
		slicingManager.set_default_profile(slicer, name, require_exists=False)

	return jsonify(_getSlicingProfileData(slicer, name, saved_profile))

@api.route("/slicing/<string:slicer>/profiles/<string:name>", methods=["DELETE"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def slicingDelSlicerProfile(slicer, name):
	try:
		slicingManager.delete_profile(slicer, name)
	except UnknownSlicer:
		return make_response("Unknown slicer {slicer}".format(**locals()), 404)
	except CouldNotDeleteProfile as e:
		return make_response("Could not delete profile {profile} for slicer {slicer}: {cause}".format(profile=name, slicer=slicer, cause=str(e.cause)), 500)

	return NO_CONTENT

def _getSlicingProfilesData(slicer, require_configured=False):
	profiles = slicingManager.all_profiles(slicer, require_configured=require_configured)

	result = dict()
	for name, profile in profiles.items():
		result[name] = _getSlicingProfileData(slicer, name, profile)
	return result

def _getSlicingProfileData(slicer, name, profile):
	defaultProfiles = s().get(["slicing", "defaultProfiles"])
	result = dict(
		key=name,
		default=defaultProfiles and slicer in defaultProfiles and defaultProfiles[slicer] == name,
		resource=url_for(".slicingGetSlicerProfile", slicer=slicer, name=name, _external=True)
	)
	if profile.display_name is not None:
		result["displayName"] = profile.display_name
	if profile.description is not None:
		result["description"] = profile.description
	return result



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net> based on work by David Braam"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2013 David Braam, Gina Häußge - Released under terms of the AGPLv3 License"


import math
import os
import io
import base64
import zlib
import logging
import codecs


class Vector3D(object):
	"""
	3D vector value

	Supports addition, subtraction and multiplication with a scalar value (float, int) as well as calculating the
	length of the vector.

	Examples:

	>>> a = Vector3D(1.0, 1.0, 1.0)
	>>> b = Vector3D(4.0, 4.0, 4.0)
	>>> a + b == Vector3D(5.0, 5.0, 5.0)
	True
	>>> b - a == Vector3D(3.0, 3.0, 3.0)
	True
	>>> abs(a - b) == Vector3D(3.0, 3.0, 3.0)
	True
	>>> a * 2 == Vector3D(2.0, 2.0, 2.0)
	True
	>>> a * 2 == 2 * a
	True
	>>> a.length == math.sqrt(a.x ** 2 + a.y ** 2 + a.z ** 2)
	True
	>>> copied_a = Vector3D(a)
	>>> a == copied_a
	True
	>>> copied_a.x == a.x and copied_a.y == a.y and copied_a.z == a.z
	True
	"""

	def __init__(self, *args):
		if len(args) == 3:
			(self.x, self.y, self.z) = args

		elif len(args) == 1:
			# copy constructor
			other = args[0]
			if not isinstance(other, Vector3D):
				raise ValueError("Object to copy must be a Vector3D instance")

			self.x = other.x
			self.y = other.y
			self.z = other.z

	@property
	def length(self):
		return math.sqrt(self.x * self.x + self.y * self.y + self.z * self.z)

	def __add__(self, other):
		try:
			if len(other) == 3:
				return Vector3D(self.x + other[0],
				                self.y + other[1],
				                self.z + other[2])
		except TypeError:
			# doesn't look like a 3-tuple
			pass

		try:
			return Vector3D(self.x + other.x,
			                self.y + other.y,
			                self.z + other.z)
		except AttributeError:
			# also doesn't look like a Vector3D
			pass

		raise TypeError("other must be a Vector3D instance or a list or tuple of length 3")

	def __sub__(self, other):
		try:
			if len(other) == 3:
				return Vector3D(self.x - other[0],
				                self.y - other[1],
				                self.z - other[2])
		except TypeError:
			# doesn't look like a 3-tuple
			pass

		try:
			return Vector3D(self.x - other.x,
			                self.y - other.y,
			                self.z - other.z)
		except AttributeError:
			# also doesn't look like a Vector3D
			pass

		raise TypeError("other must be a Vector3D instance or a list or tuple of length 3")

	def __mul__(self, other):
		try:
			return Vector3D(self.x * other,
			                self.y * other,
			                self.z * other)
		except TypeError:
			# doesn't look like a scalar
			pass

		raise ValueError("other must be a float or int value")

	def __rmul__(self, other):
		return self.__mul__(other)

	def __abs__(self):
		return Vector3D(abs(self.x), abs(self.y), abs(self.z))

	def __eq__(self, other):
		if not isinstance(other, Vector3D):
			return False
		return self.x == other.x and self.y == other.y and self.z == other.z

	def __str__(self):
		return "Vector3D(x={}, y={}, z={}, length={})".format(self.x, self.y, self.z, self.length)


class MinMax3D(object):
	"""
	Tracks minimum and maximum of recorded values

	Examples:

	>>> minmax = MinMax3D()
	>>> minmax.record(Vector3D(2.0, 2.0, 2.0))
	>>> minmax.min.x == 2.0 == minmax.max.x and minmax.min.y == 2.0 == minmax.max.y and minmax.min.z == 2.0 == minmax.max.z
	True
	>>> minmax.record(Vector3D(1.0, 2.0, 3.0))
	>>> minmax.min.x == 1.0 and minmax.min.y == 2.0 and minmax.min.z == 2.0
	True
	>>> minmax.max.x == 2.0 and minmax.max.y == 2.0 and minmax.max.z == 3.0
	True
	>>> minmax.size == Vector3D(1.0, 0.0, 1.0)
	True
	>>> empty = MinMax3D()
	>>> empty.size == Vector3D(0.0, 0.0, 0.0)
	True
	"""

	def __init__(self):
		self.min = Vector3D(float("inf"), float("inf"), float("inf"))
		self.max = Vector3D(-float("inf"), -float("inf"), -float("inf"))

	def record(self, coordinate):
		"""
		Records the coordinate, storing the min and max values.

		The input vector components must not be None.
		"""
		self.min.x = min(self.min.x, coordinate.x)
		self.min.y = min(self.min.y, coordinate.y)
		self.min.z = min(self.min.z, coordinate.z)
		self.max.x = max(self.max.x, coordinate.x)
		self.max.y = max(self.max.y, coordinate.y)
		self.max.z = max(self.max.z, coordinate.z)


	@property
	def size(self):
		result = Vector3D()
		for c in "xyz":
			min = getattr(self.min, c)
			max = getattr(self.max, c)
			value = abs(max - min) if max >= min else 0.0
			setattr(result, c, value)
		return result


class AnalysisAborted(Exception):
	def __init__(self, reenqueue=True, *args, **kwargs):
		self.reenqueue = reenqueue
		Exception.__init__(self, *args, **kwargs)


class gcode(object):
	def __init__(self, progress_callback=None):
		self._logger = logging.getLogger(__name__)
		self.layerList = None
		self.extrusionAmount = [0]
		self.extrusionVolume = [0]
		self.totalMoveTimeMinute = 0
		self.filename = None
		self._abort = False
		self._reenqueue = True
		self._filamentDiameter = 0
		self._minMax = MinMax3D()
		self._progress_callback = progress_callback

	@property
	def dimensions(self):
		size = self._minMax.size
		return dict(width=size.x,
		            depth=size.y,
		            height=size.z)

	@property
	def printing_area(self):
		return dict(minX=None if math.isinf(self._minMax.min.x) else self._minMax.min.x,
		            minY=None if math.isinf(self._minMax.min.y) else self._minMax.min.y,
		            minZ=None if math.isinf(self._minMax.min.z) else self._minMax.min.z,
		            maxX=None if math.isinf(self._minMax.max.x) else self._minMax.max.x,
		            maxY=None if math.isinf(self._minMax.max.y) else self._minMax.max.y,
		            maxZ=None if math.isinf(self._minMax.max.z) else self._minMax.max.z)

	def load(self, filename, throttle=None, speedx=6000, speedy=6000, offsets=None, max_extruders=10, g90_extruder=False):
		if os.path.isfile(filename):
			self.filename = filename
			self._fileSize = os.stat(filename).st_size

			with codecs.open(filename, encoding="utf-8", errors="replace") as f:
				self._load(f, throttle=throttle, speedx=speedx, speedy=speedy, offsets=offsets, max_extruders=max_extruders, g90_extruder=g90_extruder)

	def abort(self, reenqueue=True):
		self._abort = True
		self._reenqueue = reenqueue

	def _load(self, gcodeFile, throttle=None, speedx=6000, speedy=6000, offsets=None, max_extruders=10, g90_extruder=False):
		lineNo = 0
		readBytes = 0
		pos = Vector3D(0.0, 0.0, 0.0)
		currentE = [0.0]
		totalExtrusion = [0.0]
		maxExtrusion = [0.0]
		currentExtruder = 0
		totalMoveTimeMinute = 0.0
		relativeE = False
		relativeMode = False
		duplicationMode = False
		scale = 1.0
		fwretractTime = 0
		fwretractDist = 0
		fwrecoverTime = 0
		feedrate = min(speedx, speedy)
		if feedrate == 0:
			# some somewhat sane default if axes speeds are insane...
			feedrate = 2000

		if offsets is None or not isinstance(offsets, (list, tuple)):
			offsets = []
		if len(offsets) < max_extruders:
			offsets += [(0, 0)] * (max_extruders - len(offsets))

		for line in gcodeFile:
			if self._abort:
				raise AnalysisAborted(reenqueue=self._reenqueue)
			lineNo += 1
			readBytes += len(line.encode("utf-8"))

			if isinstance(gcodeFile, (io.IOBase, codecs.StreamReaderWriter)):
				percentage = float(readBytes) / float(self._fileSize)
			elif isinstance(gcodeFile, (list)):
				percentage = float(lineNo) / float(len(gcodeFile))
			else:
				percentage = None

			try:
				if self._progress_callback is not None and (lineNo % 1000 == 0) and percentage is not None:
					self._progress_callback(percentage)
			except Exception as exc:
				self._logger.debug("Progress callback %r error: %s", self._progress_callback, exc)

			if ';' in line:
				comment = line[line.find(';')+1:].strip()
				if comment.startswith("filament_diameter"):
					# Slic3r
					filamentValue = comment.split("=", 1)[1].strip()
					try:
						self._filamentDiameter = float(filamentValue)
					except ValueError:
						try:
							self._filamentDiameter = float(filamentValue.split(",")[0].strip())
						except ValueError:
							self._filamentDiameter = 0.0
				elif comment.startswith("CURA_PROFILE_STRING") or comment.startswith("CURA_OCTO_PROFILE_STRING"):
					# Cura 15.04.* & OctoPrint Cura plugin
					if comment.startswith("CURA_PROFILE_STRING"):
						prefix = "CURA_PROFILE_STRING:"
					else:
						prefix = "CURA_OCTO_PROFILE_STRING:"

					curaOptions = self._parseCuraProfileString(comment, prefix)
					if "filament_diameter" in curaOptions:
						try:
							self._filamentDiameter = float(curaOptions["filament_diameter"])
						except ValueError:
							self._filamentDiameter = 0.0
				elif comment.startswith("filamentDiameter,"):
					# Simplify3D
					filamentValue = comment.split(",", 1)[1].strip()
					try:
						self._filamentDiameter = float(filamentValue)
					except ValueError:
						self._filamentDiameter = 0.0
				line = line[0:line.find(';')]

			G = getCodeInt(line, 'G')
			M = getCodeInt(line, 'M')
			T = getCodeInt(line, 'T')

			if G is not None:
				if G == 0 or G == 1:	#Move
					x = getCodeFloat(line, 'X')
					y = getCodeFloat(line, 'Y')
					z = getCodeFloat(line, 'Z')
					e = getCodeFloat(line, 'E')
					f = getCodeFloat(line, 'F')

					if x is not None or y is not None or z is not None:
						# this is a move
						move = True
					else:
						# print head stays on position
						move = False

					oldPos = pos

					# Use new coordinates if provided. If not provided, use prior coordinates (minus tool offset)
					# in absolute and 0.0 in relative mode.
					newPos = Vector3D(x if x is not None else (0.0 if relativeMode else pos.x),
					                  y if y is not None else (0.0 if relativeMode else pos.y),
					                  z if z is not None else (0.0 if relativeMode else pos.z))

					if relativeMode:
						# Relative mode: scale and add to current position
						pos += newPos * scale
					else:
						# Absolute mode: scale coordinates and apply tool offsets
						pos = newPos * scale

					if f is not None and f != 0:
						feedrate = f

					if e is not None:
						if relativeMode or relativeE:
							# e is already relative, nothing to do
							pass
						else:
							e -= currentE[currentExtruder]

						# If move with extrusion, calculate new min/max coordinates of model
						if e > 0.0 and move:
							# extrusion and move -> oldPos & pos relevant for print area & dimensions
							self._minMax.record(oldPos)
							self._minMax.record(pos)

						totalExtrusion[currentExtruder] += e
						currentE[currentExtruder] += e
						maxExtrusion[currentExtruder] = max(maxExtrusion[currentExtruder],
						                                    totalExtrusion[currentExtruder])

						if currentExtruder == 0 and len(currentE) > 1 and duplicationMode:
							# Copy first extruder length to other extruders
							for i in range(1, len(currentE)):
								totalExtrusion[i] += e
								currentE[i] += e
								maxExtrusion[i] = max(maxExtrusion[i], totalExtrusion[i])
					else:
						e = 0.0

					# move time in x, y, z, will be 0 if no movement happened
					moveTimeXYZ = abs((oldPos - pos).length / feedrate)

					# time needed for extruding, will be 0 if no extrusion happened
					extrudeTime = abs(e / feedrate)

					# time to add is maximum of both
					totalMoveTimeMinute += max(moveTimeXYZ, extrudeTime)

				elif G == 4:	#Delay
					S = getCodeFloat(line, 'S')
					if S is not None:
						totalMoveTimeMinute += S / 60.0
					P = getCodeFloat(line, 'P')
					if P is not None:
						totalMoveTimeMinute += P / 60.0 / 1000.0
				elif G == 10:   #Firmware retract
					totalMoveTimeMinute += fwretractTime
				elif G == 11:   #Firmware retract recover
					totalMoveTimeMinute += fwrecoverTime
				elif G == 20:	#Units are inches
					scale = 25.4
				elif G == 21:	#Units are mm
					scale = 1.0
				elif G == 28:	#Home
					x = getCodeFloat(line, 'X')
					y = getCodeFloat(line, 'Y')
					z = getCodeFloat(line, 'Z')
					center = Vector3D(0.0, 0.0, 0.0)
					if x is None and y is None and z is None:
						pos = center
					else:
						pos = Vector3D(pos)
						if x is not None:
							pos.x = center.x
						if y is not None:
							pos.y = center.y
						if z is not None:
							pos.z = center.z
				elif G == 90:	#Absolute position
					relativeMode = False
					if g90_extruder:
						relativeE = False
				elif G == 91:	#Relative position
					relativeMode = True
					if g90_extruder:
						relativeE = True
				elif G == 92:
					x = getCodeFloat(line, 'X')
					y = getCodeFloat(line, 'Y')
					z = getCodeFloat(line, 'Z')
					e = getCodeFloat(line, 'E')

					if e is None and x is None and y is None and z is None:
						# no parameters, set all axis to 0
						currentE[currentExtruder] = 0.0
						pos.x = 0.0
						pos.y = 0.0
						pos.z = 0.0
					else:
						# some parameters set, only set provided axes
						if e is not None:
							currentE[currentExtruder] = e
						if x is not None:
							pos.x = x
						if y is not None:
							pos.y = y
						if z is not None:
							pos.z = z

			elif M is not None:
				if M == 82:   #Absolute E
					relativeE = False
				elif M == 83:   #Relative E
					relativeE = True
				elif M == 207 or M == 208: #Firmware retract settings
					s = getCodeFloat(line, 'S')
					f = getCodeFloat(line, 'F')
					if s is not None and f is not None:
						if M == 207:
							fwretractTime = s / f
							fwretractDist = s
						else:
							fwrecoverTime = (fwretractDist + s) / f
				elif M == 605:	#Duplication/Mirroring mode
					s = getCodeInt(line, 'S')
					if s in [2, 4, 5, 6]:
						# Duplication / Mirroring mode selected. Printer firmware copies extrusion commands
 						# from first extruder to all other extruders
 						duplicationMode = True
					else:
						duplicationMode = False

			elif T is not None:
				if T > max_extruders:
					self._logger.warning("GCODE tried to select tool %d, that looks wrong, ignoring for GCODE analysis" % T)
				elif T == currentExtruder:
					pass
				else:
					pos.x -= offsets[currentExtruder][0] if currentExtruder < len(offsets) else 0
					pos.y -= offsets[currentExtruder][1] if currentExtruder < len(offsets) else 0

					currentExtruder = T

					pos.x += offsets[currentExtruder][0] if currentExtruder < len(offsets) else 0
					pos.y += offsets[currentExtruder][1] if currentExtruder < len(offsets) else 0

					if len(currentE) <= currentExtruder:
						for i in range(len(currentE), currentExtruder + 1):
							currentE.append(0.0)
					if len(maxExtrusion) <= currentExtruder:
						for i in range(len(maxExtrusion), currentExtruder + 1):
							maxExtrusion.append(0.0)
					if len(totalExtrusion) <= currentExtruder:
						for i in range(len(totalExtrusion), currentExtruder + 1):
							totalExtrusion.append(0.0)

			if throttle is not None:
				throttle(lineNo, readBytes)
		if self._progress_callback is not None:
			self._progress_callback(100.0)

		self.extrusionAmount = maxExtrusion
		self.extrusionVolume = [0] * len(maxExtrusion)
		for i in range(len(maxExtrusion)):
			radius = self._filamentDiameter / 2
			self.extrusionVolume[i] = (self.extrusionAmount[i] * (math.pi * radius * radius)) / 1000
		self.totalMoveTimeMinute = totalMoveTimeMinute

	def _parseCuraProfileString(self, comment, prefix):
		return {key: value for (key, value) in map(lambda x: x.split("=", 1), zlib.decompress(base64.b64decode(comment[len(prefix):])).split("\b"))}

	def get_result(self):
		return dict(total_time=self.totalMoveTimeMinute,
		            extrusion_length=self.extrusionAmount,
		            extrusion_volume=self.extrusionVolume,
		            dimensions=self.dimensions,
		            printing_area=self.printing_area)

def getCodeInt(line, code):
	return getCode(line, code, int)


def getCodeFloat(line, code):
	return getCode(line, code, float)


def getCode(line, code, c):
	n = line.find(code) + 1
	if n < 1:
		return None
	m = line.find(' ', n)
	try:
		if m < 0:
			result = c(line[n:])
		else:
			result = c(line[n:m])
	except ValueError:
		return None

	if math.isnan(result) or math.isinf(result):
		return None

	return result



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"


import collections
import frozendict

from octoprint.util import to_unicode

class JsonEncoding(object):

	encoders = collections.OrderedDict()

	@classmethod
	def add_encoder(cls, type, encoder):
		cls.encoders[type] = encoder

	@classmethod
	def remove_encoder(cls, type):
		try:
			del cls.encoders[type]
		except KeyError:
			pass

	@classmethod
	def encode(cls, obj):
		for type, encoder in cls.encoders.items():
			if isinstance(obj, type):
				return encoder(obj)
		raise TypeError

JsonEncoding.add_encoder(frozendict.frozendict, lambda obj: dict(obj))
JsonEncoding.add_encoder(bytes, lambda obj: to_unicode(obj))
